import json

import requests
from requests.exceptions import HTTPError

from apigee import APIGEE_ADMIN_API_URL, auth, console
from apigee.sharedflows.serializer import SharedflowsSerializer
from apigee.utils import add_to_dict_if_exists

GET_A_LIST_OF_SHARED_FLOWS_PATH = '{api_url}/v1/organizations/{org}/sharedflows'
IMPORT_A_SHARED_FLOW_PATH = '{api_url}/v1/organizations/{org}/sharedflows'
# EXPORT_A_SHARED_FLOW_PATH = ''
GET_A_SHARED_FLOW_PATH = '{api_url}/v1/organizations/{org}/sharedflows/{shared_flow_name}'
DEPLOY_A_SHARED_FLOW_PATH = '{api_url}/v1/organizations/{org}/environments/{environment}/sharedflows/{shared_flow_name}/revisions/{revision_number}/deployments'
UNDEPLOY_SHARED_FLOW_REVISIONS_IN_ENVIRONMENT_PATH = ''
UNDEPLOY_A_SHARED_FLOW_PATH = '{api_url}/v1/organizations/{org}/environments/{environment}/sharedflows/{shared_flow_name}/revisions/{revision_number}/deployments'
GET_DEPLOYMENT_ENVIRONMENTS_FOR_SHARED_FLOWS_PATH = '{api_url}/v1/organizations/{org}/sharedflows/{shared_flow_name}/revisions/{revision_number}/deployments'
DELETE_A_SHARED_FLOW_PATH = ''
# ATTACH_A_SHARED_FLOW_TO_A_FLOW_HOOK_PATH = ''
# DETACHES_A_SHARED_FLOW_FROM_A_FLOW_HOOK_PATH = ''
GET_THE_SHARED_FLOW_ATTACHED_TO_A_FLOW_HOOK_PATH = (
    '{api_url}/v1/organizations/{org}/environments/{environment}/flowhooks/{flow_hook}'
)
GET_SHARED_FLOW_DEPLOYMENTS_PATH = (
    '{api_url}/v1/organizations/{org}/sharedflows/{shared_flow_name}/deployments'
)
GET_SHARED_FLOW_REVISIONS_PATH = (
    '{api_url}/v1/organizations/{org}/sharedflows/{shared_flow_name}/revisions'
)


class Sharedflows:
    def __init__(self, auth, org_name):
        self._auth = auth
        self._org_name = org_name

    def __call__(self):
        pass

    @property
    def auth(self):
        return self._auth

    @auth.setter
    def auth(self, value):
        self._auth = value

    @property
    def org_name(self):
        return self._org_name

    @org_name.setter
    def org_name(self, value):
        self._org_name = value

    def get_a_list_of_shared_flows(self, prefix=None):
        uri = GET_A_LIST_OF_SHARED_FLOWS_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name
        )
        hdrs = auth.set_header(self._auth, {'Accept': 'application/json'})
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return SharedflowsSerializer().serialize_details(resp, 'json', prefix=prefix)

    def import_a_shared_flow(self, shared_flow_file, shared_flow_name):
        uri = IMPORT_A_SHARED_FLOW_PATH.format(api_url=APIGEE_ADMIN_API_URL, org=self._org_name)
        hdrs = auth.set_header(
            self._auth, {'Accept': 'application/json', 'Content-Type': 'multipart/form-data'}
        )
        params = {'action': 'import', 'name': shared_flow_name}
        with open(shared_flow_file, 'rb') as f:
            resp = requests.post(
                uri, headers=hdrs, files={'file': ('sharedflow.zip', f)}, params=params
            )
        resp.raise_for_status()
        return resp

    # def export_a_shared_flow(self, shared_flow_name, revision_number):
    #     pass

    def get_a_shared_flow(self, shared_flow_name):
        uri = GET_A_SHARED_FLOW_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, shared_flow_name=shared_flow_name
        )
        hdrs = auth.set_header(self._auth, {'Accept': 'application/json'})
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def deploy_a_shared_flow(
        self,
        environment,
        shared_flow_name,
        revision_number,
        override=False,
        delay=0,
        shared_flow_file=None,
    ):
        do_deployments_exist = False
        try:
            if self.get_shared_flow_revisions(shared_flow_name):
                do_deployments_exist = True
        except HTTPError as e:
            if e.response.status_code != 404:
                raise e
        if shared_flow_file:
            revision_number = int(
                self.import_a_shared_flow(shared_flow_file, shared_flow_name).json()['revision']
            )
        uri = DEPLOY_A_SHARED_FLOW_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            environment=environment,
            shared_flow_name=shared_flow_name,
            revision_number=revision_number,
        )
        options_dict = {'override': 'true' if override else 'false', 'delay': f'{delay}'}
        params = add_to_dict_if_exists(options_dict)
        hdrs = auth.set_header(self._auth, {'Accept': 'application/json'})
        console.echo(f'Deploying revision {revision_number}... ', end='', flush=True)
        resp = requests.post(uri, headers=hdrs, params=params)
        resp.raise_for_status()
        console.echo('Done')
        if do_deployments_exist:
            # console.echo('Attempting undeployment... ')
            self.undeploy_shared_flow_revisions_in_environment(
                environment, shared_flow_name, except_revisions={revision_number}
            )
            # console.echo('Done.')
        return resp

    def undeploy_shared_flow_revisions_in_environment(
        self, environment, shared_flow_name, except_revisions=set()
    ):
        resp = self.get_shared_flow_deployments(shared_flow_name)
        for deployment in resp.json()['environment']:
            if deployment['name'] == environment:
                for detail in deployment['revision']:
                    revision_number = int(detail['name'])
                    if revision_number not in except_revisions:
                        console.echo(
                            f'Undeploying revision {revision_number}... ', end='', flush=True
                        )
                        self.undeploy_a_shared_flow(environment, shared_flow_name, revision_number)
                        console.echo('Done')
        return resp

    def undeploy_a_shared_flow(self, environment, shared_flow_name, revision_number):
        uri = UNDEPLOY_A_SHARED_FLOW_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            environment=environment,
            shared_flow_name=shared_flow_name,
            revision_number=revision_number,
        )
        hdrs = auth.set_header(self._auth, {'Accept': 'application/json'})
        resp = requests.delete(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def get_deployment_environments_for_shared_flows(self, shared_flow_name, revision_number):
        uri = GET_DEPLOYMENT_ENVIRONMENTS_FOR_SHARED_FLOWS_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            shared_flow_name=shared_flow_name,
            revision_number=revision_number,
        )
        hdrs = auth.set_header(self._auth, {'Accept': 'application/json'})
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def delete_a_shared_flow(self, shared_flow_name):
        pass

    # def attach_a_shared_flow_to_a_flow_hook(self, environment, flow_hook, request_body):
    #     pass
    #
    # def detaches_a_shared_flow_from_a_flow_hook(self, environment, flow_hook):
    #     pass

    def get_the_shared_flow_attached_to_a_flow_hook(self, environment, flow_hook):
        uri = GET_THE_SHARED_FLOW_ATTACHED_TO_A_FLOW_HOOK_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            environment=environment,
            flow_hook=flow_hook,
        )
        hdrs = auth.set_header(self._auth, {'Accept': 'application/json'})
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def get_shared_flow_deployments(self, shared_flow_name):
        uri = GET_SHARED_FLOW_DEPLOYMENTS_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, shared_flow_name=shared_flow_name
        )
        hdrs = auth.set_header(self._auth, {'Accept': 'application/json'})
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def get_shared_flow_revisions(self, shared_flow_name):
        uri = GET_SHARED_FLOW_REVISIONS_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, shared_flow_name=shared_flow_name
        )
        hdrs = auth.set_header(self._auth, {'Accept': 'application/json'})
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return resp



import configparser
import sys

import click
from click_aliases import ClickAliasedGroup

from apigee import APIGEE_CLI_CREDENTIALS_FILE, APIGEE_CLI_DIRECTORY
from apigee.utils import make_dirs


class HiddenSecret(object):
    def __init__(self, secret=''):
        self.secret = secret

    def __str__(self):
        return '*' * 16 if self.secret else ''


KEY_LIST = ('username', 'password', 'mfa_secret', 'is_token', 'zonename', 'org', 'prefix')
config = configparser.ConfigParser()
config.read(APIGEE_CLI_CREDENTIALS_FILE)
profile = 'default'

for i, arg in enumerate(sys.argv):
    if arg == '-P' or arg == '--profile':
        try:
            profile = sys.argv[i + 1]
        except IndexError:
            pass
try:
    profile_dict = dict(config._sections[profile])
    for key in KEY_LIST:
        if key not in profile_dict:
            profile_dict[key] = ''
except KeyError:
    profile_dict = {k: '' for k in KEY_LIST}

# @click.group(cls=ClickAliasedGroup)
# @click.command(aliases=['conf', 'config', 'cfg'])
@click.command(help='Configure Apigee Edge credentials.')
@click.option(
    '-u',
    '--username',
    prompt='Apigee username (email)',
    default=profile_dict['username'],
    show_default=True,
)
@click.option(
    '-p',
    '--password',
    prompt='Apigee password',
    default=lambda: HiddenSecret(profile_dict['password']),
    hide_input=True,
    show_default='hidden' if profile_dict['password'] else 'None',
)
@click.option(
    '-mfa',
    '--mfa-secret',
    prompt='Apigee MFA key (optional)',
    default=lambda: HiddenSecret(profile_dict['mfa_secret']),
    hide_input=True,
    show_default='hidden' if profile_dict['mfa_secret'] else 'None',
)
@click.option(
    '-z',
    '--zonename',
    prompt='Identity zone name (to support SAML authentication)',
    default=profile_dict['zonename'],
    show_default=True,
)
@click.option(
    '--token/--no-token',
    default=True if profile_dict['is_token'] in (True, 'True', 'true', '1') else False,
    help='specify to use oauth without MFA',
    prompt='Use OAuth, no MFA (optional)?',
    show_default=True,
)
@click.option(
    '-o',
    '--org',
    prompt='Default Apigee organization (recommended)',
    default=profile_dict['org'],
    show_default=True,
)
@click.option(
    '--prefix',
    prompt='Default team/resource prefix (optional)',
    default=profile_dict['prefix'],
    show_default=True,
)
@click.option(
    '-P',
    '--profile',
    help='name of the user profile to create/update',
    default='default',
    show_default=True,
)
def configure(username, password, mfa_secret, token, zonename, org, prefix, profile):
    if isinstance(password, HiddenSecret):
        password = password.secret
    if isinstance(mfa_secret, HiddenSecret):
        mfa_secret = mfa_secret.secret
    profile_dict['username'] = username
    profile_dict['password'] = password
    profile_dict['mfa_secret'] = mfa_secret
    profile_dict['is_token'] = token
    profile_dict['zonename'] = zonename
    profile_dict['org'] = org
    profile_dict['prefix'] = prefix
    config[profile] = {k: v for k, v in profile_dict.items() if v}
    make_dirs(APIGEE_CLI_DIRECTORY)
    with open(APIGEE_CLI_CREDENTIALS_FILE, 'w') as cf:
        config.write(cf)



import json


class AppsSerializer:
    def serialize_details(self, apps, format, prefix=None):
        resp = apps
        if format == 'text':
            return apps.text
        apps = apps.json()
        if prefix:
            apps = [app for app in apps if app.startswith(prefix)]
        if format == 'json':
            return json.dumps(apps)
        elif format == 'table':
            pass
        elif format == 'dict':
            return apps
        return resp



import click

from apigee import console
from apigee.auth import common_auth_options, gen_auth
from apigee.prefix import common_prefix_options
from apigee.references.references import References
from apigee.silent import common_silent_options
from apigee.verbose import common_verbose_options


@click.group(help='References in an organization and environment.')
def references():
    pass


def _list_all_references(
    username,
    password,
    mfa_secret,
    token,
    zonename,
    org,
    profile,
    environment,
    prefix=None,
    **kwargs
):
    return References(
        gen_auth(username, password, mfa_secret, token, zonename), org, None
    ).list_all_references(environment, prefix=prefix)


@references.command(help='List all references in an organization and environment.')
@common_auth_options
@common_prefix_options
@common_silent_options
@common_verbose_options
@click.option('-e', '--environment', help='environment', required=True)
def list(*args, **kwargs):
    console.echo(_list_all_references(*args, **kwargs))


def _get_reference(
    username, password, mfa_secret, token, zonename, org, profile, name, environment, **kwargs
):
    return (
        References(gen_auth(username, password, mfa_secret, token, zonename), org, name)
        .get_reference(environment)
        .text
    )


@references.command(help='Get reference in an organization and environment.')
@common_auth_options
@common_silent_options
@common_verbose_options
@click.option('-n', '--name', help='reference name', required=True)
@click.option('-e', '--environment', help='environment', required=True)
def get(*args, **kwargs):
    console.echo(_get_reference(*args, **kwargs))


def _delete_reference(*args, **kwargs):
    pass


@references.command(help='')
@common_auth_options
@common_silent_options
@common_verbose_options
def delete(*args, **kwargs):
    _delete_reference(*args, **kwargs)


def _create_reference(*args, **kwargs):
    pass


@references.command(help='')
@common_auth_options
@common_silent_options
@common_verbose_options
def create(*args, **kwargs):
    _create_reference(*args, **kwargs)


def _update_reference(*args, **kwargs):
    pass


@references.command(help='')
@common_auth_options
@common_silent_options
@common_verbose_options
def update(*args, **kwargs):
    _update_reference(*args, **kwargs)



import click

from apigee import console
from apigee.auth import common_auth_options, gen_auth
from apigee.developers.developers import Developers
from apigee.prefix import common_prefix_options
from apigee.silent import common_silent_options
from apigee.verbose import common_verbose_options


@click.group(
    help='Developers implement client/consumer apps and must be registered with an organization on Apigee Edge.'
)
def developers():
    pass


def _create_developer(
    username,
    password,
    mfa_secret,
    token,
    zonename,
    org,
    profile,
    name,
    first_name,
    last_name,
    user_name,
    attributes='{"attributes" : [ ]}',
    **kwargs
):
    return (
        Developers(gen_auth(username, password, mfa_secret, token, zonename), org, name)
        .create_developer(first_name, last_name, user_name, attributes=attributes)
        .text
    )


@developers.command(
    help='Creates a profile for a developer in an organization. Once created, the developer can register an app and receive an API key.'
)
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option(
    '-n',
    '--name',
    help="The developer's email. This value is used to uniquely identify the developer in Apigee Edge.",
    required=True,
)
@click.option('--first-name', help='The first name of the developer.', required=True)
@click.option('--last-name', help='The last name of the developer.', required=True)
@click.option(
    '--user-name',
    help="The developer's username. This value is not used by Apigee Edge.",
    required=True,
)
@click.option(
    '--attributes',
    default='{"attributes" : [ ]}',
    required=True,
    help='request body e.g.: \'{"attributes" : [ ]}\'',
)
def create(*args, **kwargs):
    console.echo(_create_developer(*args, **kwargs))


def _delete_developer(
    username, password, mfa_secret, token, zonename, org, profile, name, **kwargs
):
    return (
        Developers(gen_auth(username, password, mfa_secret, token, zonename), org, name)
        .delete_developer()
        .text
    )


@developers.command(
    help='Deletes a developer from an organization. All apps and API keys associated with the developer are also removed from the organization. All times in the response are UNIX times. To avoid permanently deleting developers and their artifacts, consider deactivating developers instead.'
)
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option(
    '-n',
    '--name',
    help="The developer's email. This value is used to uniquely identify the developer in Apigee Edge.",
    required=True,
)
def delete(*args, **kwargs):
    console.echo(_delete_developer(*args, **kwargs))


def _get_developer(username, password, mfa_secret, token, zonename, org, profile, name, **kwargs):
    return (
        Developers(gen_auth(username, password, mfa_secret, token, zonename), org, name)
        .get_developer()
        .text
    )


@developers.command(
    help="Returns the profile for a developer by email address or ID. All time values are UNIX time values. The profile includes the developer's email address, ID, name, and other information."
)
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option(
    '-n',
    '--name',
    help="The developer's email. This value is used to uniquely identify the developer in Apigee Edge.",
    required=True,
)
def get(*args, **kwargs):
    console.echo(_get_developer(*args, **kwargs))


def _get_developer_by_app(
    username, password, mfa_secret, token, zonename, org, profile, app_name, **kwargs
):
    return (
        Developers(gen_auth(username, password, mfa_secret, token, zonename), org, None)
        .get_developer_by_app(app_name)
        .text
    )


@developers.command(
    help='Gets the developer profile by app name. The profile retrieved is for the developer associated with the app in the organization. All time values are UNIX time values.'
)
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option('--app-name', help='the app name', required=True)
def get_by_app(*args, **kwargs):
    console.echo(_get_developer_by_app(*args, **kwargs))


def _list_developers(
    username,
    password,
    mfa_secret,
    token,
    zonename,
    org,
    profile,
    prefix=None,
    expand=False,
    count=1000,
    startkey="",
    **kwargs
):
    return Developers(
        gen_auth(username, password, mfa_secret, token, zonename), org, None
    ).list_developers(prefix=prefix, expand=expand, count=count, startkey=startkey)


@developers.command(
    help='Lists all developers in an organization by email address. This call does not list any company developers who are a part of the designated organization.'
)
@common_auth_options
@common_verbose_options
@common_silent_options
@common_prefix_options
@click.option(
    '--expand/--no-expand',
    default=False,
    help='Set to true to list developers exanded with details.',
)
@click.option(
    '--count',
    type=click.INT,
    default=1000,
    show_default=True,
    help='Limits the list to the number you specify. Use with the startKey parameter to provide more targeted filtering. The limit is 1000.',
)
@click.option(
    '--startkey',
    default="",
    show_default=True,
    help='To filter the keys that are returned, enter the email of a developer that the list will start with.',
)
def list(*args, **kwargs):
    console.echo(_list_developers(*args, **kwargs))


def _set_developer_status(
    username, password, mfa_secret, token, zonename, org, profile, name, action, **kwargs
):
    return (
        Developers(gen_auth(username, password, mfa_secret, token, zonename), org, name)
        .set_developer_status(action)
        .text
    )


@developers.command(
    help="Sets a developer's status to active or inactive for a specific organization. Run this API for each organization where you want to change the developer's status."
)
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option(
    '-n',
    '--name',
    help="The developer's email. This value is used to uniquely identify the developer in Apigee Edge.",
    required=True,
)
@click.option(
    '--action',
    type=click.Choice(['active', 'inactive'], case_sensitive=False),
    required=True,
    show_default=True,
)
def set_status(*args, **kwargs):
    console.echo(_set_developer_status(*args, **kwargs))


def _update_developer(
    username, password, mfa_secret, token, zonename, org, profile, name, body, **kwargs
):
    return (
        Developers(gen_auth(username, password, mfa_secret, token, zonename), org, name)
        .update_developer(body)
        .text
    )


@developers.command(
    help='Update an existing developer profile. To add new values or update existing values, submit the new or updated portion of the developer profile along with the rest of the developer profile, even if no values are changing. To delete attributes from a developer profile, submit the entire profile without the attributes that you want to delete.'
)
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option(
    '-n',
    '--name',
    help="The developer's email. This value is used to uniquely identify the developer in Apigee Edge.",
    required=True,
)
@click.option('-b', '--body', help='request body', required=True)
def update(*args, **kwargs):
    console.echo(_update_developer(*args, **kwargs))


# def _get_developer_attribute(username, password, mfa_secret, token, zonename, org, profile, attribute_name, **kwargs):
#     pass


def _update_a_developer_attribute(
    username,
    password,
    mfa_secret,
    token,
    zonename,
    org,
    profile,
    name,
    attribute_name,
    updated_value,
    **kwargs
):
    return (
        Developers(gen_auth(username, password, mfa_secret, token, zonename), org, name)
        .update_a_developer_attribute(attribute_name, updated_value)
        .text
    )


@developers.command(help='Updates the value of a developer attribute.')
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option(
    '-n',
    '--name',
    help="The developer's email. This value is used to uniquely identify the developer in Apigee Edge.",
    required=True,
)
@click.option('--attribute-name', help='attribute name', required=True)
@click.option('--updated-value', help='updated value', required=True)
def update_attr(*args, **kwargs):
    console.echo(_update_a_developer_attribute(*args, **kwargs))


def _delete_developer_attribute(
    username, password, mfa_secret, token, zonename, org, profile, name, attribute_name, **kwargs
):
    return (
        Developers(gen_auth(username, password, mfa_secret, token, zonename), org, name)
        .delete_developer_attribute(attribute_name)
        .text
    )


@developers.command(help='Deletes a developer attribute.')
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option(
    '-n',
    '--name',
    help="The developer's email. This value is used to uniquely identify the developer in Apigee Edge.",
    required=True,
)
@click.option('--attribute-name', help='attribute name', required=True)
def delete_attr(*args, **kwargs):
    console.echo(_delete_developer_attribute(*args, **kwargs))


def _get_all_developer_attributes(
    username, password, mfa_secret, token, zonename, org, profile, name, **kwargs
):
    return (
        Developers(gen_auth(username, password, mfa_secret, token, zonename), org, name)
        .get_all_developer_attributes()
        .text
    )


@developers.command(help='Returns a list of all developer attributes.')
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option(
    '-n',
    '--name',
    help="The developer's email. This value is used to uniquely identify the developer in Apigee Edge.",
    required=True,
)
def get_attrs(*args, **kwargs):
    console.echo(_get_all_developer_attributes(*args, **kwargs))


def _update_all_developer_attributes(
    username, password, mfa_secret, token, zonename, org, profile, name, body, **kwargs
):
    return (
        Developers(gen_auth(username, password, mfa_secret, token, zonename), org, name)
        .update_all_developer_attributes(body)
        .text
    )


@developers.command(
    help='Updates or creates developer attributes. This API replaces the current list of attributes with the attributes specified in the request body. This lets you update existing attributes, add new attributes, or delete existing attributes by omitting them from the request body.'
)
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option(
    '-n',
    '--name',
    help="The developer's email. This value is used to uniquely identify the developer in Apigee Edge.",
    required=True,
)
@click.option('-b', '--body', help='request body', required=True)
def update_all_attrs(*args, **kwargs):
    console.echo(_update_all_developer_attributes(*args, **kwargs))



APIS = {
    'apis',
    'keyvaluemaps',
    'targetservers',
    'caches',
    'developers',
    'apiproducts',
    'apps',
    'userroles',
}


class Struct:
    def __init__(self, **entries):
        self.__dict__.update(entries)


def empty_snapshot():
    return Struct(
        apis={},
        keyvaluemaps={},
        targetservers={},
        caches={},
        developers=[],
        apps={},
        apiproducts=[],
        userroles=[],
    )



import configparser
import json
import os
import shutil
import stat
import sys
from os import path
from pathlib import Path

import click
from click_option_group import MutuallyExclusiveOptionGroup, optgroup

from apigee import (APIGEE_CLI_PLUGINS_CONFIG_FILE,
                    APIGEE_CLI_PLUGINS_DIRECTORY, APIGEE_CLI_PLUGINS_PATH,
                    console)
from apigee.silent import common_silent_options
from apigee.utils import (is_dir, is_file, make_dirs, read_file,
                          run_func_on_dir_files, touch)
from apigee.verbose import common_verbose_options

is_git_installed = False
plugins_command_help = '[Experimental] Simple plugins manager for distributing commands.'

try:
    import git
    from git import Git, Repo

    is_git_installed = True
except ImportError:
    plugins_command_help = (
        '[Unavailable - Git not found] Simple plugins manager for distributing commands.'
    )


def exit_if_git_not_installed():
    if not is_git_installed:
        sys.exit(0)


@click.group(help=plugins_command_help)
def plugins():
    pass


def init():
    make_dirs(APIGEE_CLI_PLUGINS_DIRECTORY)
    touch(APIGEE_CLI_PLUGINS_PATH)
    touch(APIGEE_CLI_PLUGINS_CONFIG_FILE)


def load_config(config_file=APIGEE_CLI_PLUGINS_CONFIG_FILE):
    config = configparser.ConfigParser(allow_no_value=True)
    config.read(config_file)
    return config


def save_config(plugins_file, config_file, section='sources'):
    init()
    config = load_config()
    if not config._sections:
        return


def clone_all(section='sources'):
    init()
    config = load_config()
    if not config._sections:
        return
    sources = dict(config._sections[section])
    for name, uri in sources.items():
        dest = Path(APIGEE_CLI_PLUGINS_DIRECTORY) / name
        if is_dir(dest):
            continue
        try:
            console.echo(f'Installing {name}... ', end='', flush=True)
            Repo.clone_from(uri, dest)
            console.echo('Done')
        except Exception as e:
            console.echo(e)


def pull_all():
    def _func(path):
        if not is_dir(path):
            return
        console.echo(f'Updating {Path(path).stem}... ', end='', flush=True)
        repo = Repo(path)
        if repo.bare:
            return
        try:
            repo.remotes['origin'].pull()
            console.echo('Done')
        except Exception as e:
            console.echo(e)

    return run_func_on_dir_files(APIGEE_CLI_PLUGINS_DIRECTORY, _func, glob='[!.][!__]*')


@plugins.command(help='Edit config file manually.')
@common_silent_options
@common_verbose_options
@click.option(
    '-a/-A',
    '--apply-changes/--no-apply-changes',
    default=False,
    help='Install plugins from new sources after exiting the editor.',
    show_default=True,
)
def configure(silent, verbose, apply_changes):
    exit_if_git_not_installed()
    init()
    click.edit(filename=APIGEE_CLI_PLUGINS_CONFIG_FILE)
    if apply_changes:
        clone_all()
        prune_all()
    else:
        console.echo('\n  Run `apigee plugins update` to apply any changes,')
        console.echo('    or rerun `apigee plugins configure` with `-a`')
        console.echo('    to apply changes automatically.\n')


def install():
    pass


@plugins.command(help='Update or install plugins.')
@common_silent_options
@common_verbose_options
def update(silent, verbose, section='sources'):
    exit_if_git_not_installed()
    clone_all()
    pull_all()


@plugins.command(help='Show plugins information.')
@common_silent_options
@common_verbose_options
@click.option('-n', '--name', help='name of the plugins package')
@optgroup.group('Filter options', cls=MutuallyExclusiveOptionGroup, help='The filter options')
@optgroup.option(
    '--show-commit-only/--no-show-commit-only',
    default=False,
    help='only print latest Git commit log',
)
@optgroup.option(
    '--show-dependencies-only/--no-show-dependencies-only',
    default=False,
    help='only print list of required packages',
)
def show(
    silent, verbose, name, section='sources', show_commit_only=False, show_dependencies_only=False
):
    if not name:
        config = load_config()
        if not config._sections:
            return
        sources = dict(config._sections[section])
        for name, uri in sources.items():
            console.echo(f'{name}: {uri}')
        return
    plugins_info_file = Path(APIGEE_CLI_PLUGINS_DIRECTORY) / name / 'apigee-cli.info'
    if not is_file(plugins_info_file):
        return
    plugins_info = read_file(plugins_info_file, type='json')
    if show_commit_only:
        exit_if_git_not_installed()
        console.echo(
            Repo(Path(APIGEE_CLI_PLUGINS_DIRECTORY) / name).git.log(
                '--pretty=format:%Cred%h%Creset -%C(yellow)%d%Creset %s %Cgreen(%cr) %C(bold blue)<%an>%Creset',
                '-1',
            )
        )
        return
    if show_dependencies_only:
        if plugins_info.get('Requires'):
            console.echo(plugins_info.get('Requires'))
        return
    for k, v in plugins_info.items():
        console.echo(f'{k}: {v}')


def info():
    pass


def chmod_directory(directory, mode):
    """https://stackoverflow.com/a/58878271"""
    for root, dirs, files in os.walk(directory):
        for dir in dirs:
            os.chmod(path.join(root, dir), mode)
        for file in files:
            os.chmod(path.join(root, file), mode)


def prune_all(section='sources'):
    init()
    config = load_config()
    if not config._sections:
        return
    sources = dict(config._sections[section])

    def _func(path):
        if not is_dir(path):
            return
        name = Path(path).stem
        if name in sources.keys():
            return
        console.echo(f'Removing {name}... ', end='', flush=True)
        plugin_directory = Path(APIGEE_CLI_PLUGINS_DIRECTORY) / name
        try:
            chmod_directory(str(Path(plugin_directory) / '.git'), stat.S_IRWXU)
            shutil.rmtree(plugin_directory)
            console.echo('Done')
        except Exception as e:
            console.echo(e)

    return run_func_on_dir_files(APIGEE_CLI_PLUGINS_DIRECTORY, _func, glob='[!.][!__]*')


@plugins.command(help='Prune plugins with removed sources.')
@common_silent_options
@common_verbose_options
def prune(silent, verbose, section='sources'):
    exit_if_git_not_installed()
    prune_all()


def uninstall():
    pass


def clean():
    pass



import base64
import binascii
import configparser
import inspect
import json
import os
import sys
import time
import urllib.error
import urllib.parse
import urllib.request
import webbrowser

import click
import jwt
import pyotp
import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import ConnectionError
from requests.packages.urllib3.util.retry import Retry

from apigee import (APIGEE_CLI_ACCESS_TOKEN_FILE, APIGEE_CLI_CREDENTIALS_FILE,
                    APIGEE_CLI_DIRECTORY, APIGEE_CLI_IS_MACHINE_USER,
                    APIGEE_OAUTH_URL, APIGEE_SAML_LOGIN_URL,
                    APIGEE_ZONENAME_OAUTH_URL, console)
from apigee.cls import AliasedGroup
# from apigee.prefix import common_prefix_options
from apigee.prefix import auth_with_prefix as with_prefix
from apigee.silent import common_silent_options
from apigee.types import Struct
from apigee.utils import make_dirs
from apigee.verbose import common_verbose_options


def attach_username_option(func, profile):
    username = get_credential(profile, 'username')
    username_envvar = os.environ.get(f'APIGEE_USERNAME', '')
    if username:
        func = click.option('-u', '--username', default=username, show_default='current username')(
            func
        )
    elif username_envvar:
        func = click.option(
            '-u', '--username', default=username_envvar, show_default='current username'
        )(func)
    else:
        func = click.option('-u', '--username', required=True)(func)
    return func


def attach_password_option(func, profile):
    password = get_credential(profile, 'password')
    password_envvar = os.environ.get(f'APIGEE_PASSWORD', '')
    if password:
        func = click.option('-p', '--password', default=password, show_default='current password')(
            func
        )
    elif password_envvar:
        func = click.option(
            '-p', '--password', default=password_envvar, show_default='current password'
        )(func)
    else:
        func = click.option('-p', '--password', required=True)(func)
    return func


def attach_mfa_secret_option(func, profile):
    mfa_secret = get_credential(profile, 'mfa_secret')
    mfa_envvar = os.environ.get(f'APIGEE_MFA_SECRET', '')
    if mfa_secret:
        func = click.option(
            '-mfa', '--mfa-secret', default=mfa_secret, show_default='current mfa key'
        )(func)
    elif mfa_envvar:
        func = click.option(
            '-mfa', '--mfa-secret', default=mfa_envvar, show_default='current mfa key'
        )(func)
    else:
        func = click.option('-mfa', '--mfa-secret')(func)
    return func


def attach_is_token_option(func, profile):
    is_token = get_credential(profile, 'is_token')
    is_token_envvar = os.environ.get(f'APIGEE_IS_TOKEN', '')
    if is_token in (True, 'True', 'true', '1'):
        func = click.option(
            '--token/--no-token',
            default=is_token,
            show_default=True,
            help='specify to use oauth without MFA',
        )(func)
    elif is_token_envvar in (True, 'True', 'true', '1'):
        func = click.option(
            '--token/--no-token',
            default=is_token_envvar,
            show_default=True,
            help='specify to use oauth without MFA',
        )(func)
    else:
        func = click.option(
            '--token/--no-token',
            default=False,
            show_default=True,
            help='specify to use oauth without MFA',
        )(func)
    return func


def attach_zonename_option(func, profile):
    zonename = get_credential(profile, 'zonename')
    zonename_envvar = os.environ.get('APIGEE_ZONENAME', '')
    if zonename:
        func = click.option(
            '-z', '--zonename', default=zonename, show_default='current identity zone name'
        )(func)
    elif zonename_envvar:
        func = click.option(
            '-z', '--zonename', default=zonename_envvar, show_default='current identity zone name'
        )(func)
    else:
        func = click.option('-z', '--zonename', help='identity zone name')(func)
    return func


def attach_org_option(func, profile):
    org = get_credential(profile, 'org')
    org_envvar = os.environ.get(f'APIGEE_ORG', '')
    if org:
        func = click.option('-o', '--org', default=org, show_default='current org')(func)
    elif org_envvar:
        func = click.option('-o', '--org', default=org_envvar, show_default='current org')(func)
    else:
        func = click.option('-o', '--org', required=True)(func)
    return func


def common_auth_options(func):
    profile = 'default'
    for i, arg in enumerate(sys.argv):
        if arg == '-P' or arg == '--profile':
            try:
                profile = sys.argv[i + 1]
            except IndexError:
                pass
    attach_username_option(func, profile)
    attach_password_option(func, profile)
    attach_mfa_secret_option(func, profile)
    attach_is_token_option(func, profile)
    attach_zonename_option(func, profile)
    attach_org_option(func, profile)
    func = click.option(
        '-P',
        '--profile',
        help='name of the user profile to authenticate with',
        default=profile,
        show_default=True,
    )(func)
    return func


def gen_auth(username=None, password=None, mfa_secret=None, token=None, zonename=None):
    return Struct(
        username=username, password=password, mfa_secret=mfa_secret, token=token, zonename=zonename
    )


def get_access_token_for_token(auth, username, password, oauth_url, post_headers, session):
    if auth.zonename:
        oauth_url = APIGEE_ZONENAME_OAUTH_URL.format(zonename=auth.zonename)
    post_body = f'username={urllib.parse.quote(username)}&password={urllib.parse.quote(password)}&grant_type=password&response_type=token'
    try:
        return session.post(f'{oauth_url}', headers=post_headers, data=post_body)
    except ConnectionError as ce:
        console.echo(ce)


def get_access_token_for_mfa(auth, username, password, oauth_url, post_headers, session):
    mfa_secret = auth.mfa_secret
    totp = pyotp.TOTP(mfa_secret)
    try:
        totp.now()
    except binascii.Error as e:
        sys.exit(f'{type(e).__name__}: {e}: Not a valid MFA key')
    post_body = f'username={urllib.parse.quote(username)}&password={urllib.parse.quote(password)}&grant_type=password'
    try:
        response_post = session.post(
            f'{oauth_url}?mfa_token={totp.now()}', headers=post_headers, data=post_body
        )
    except ConnectionError as ce:
        console.echo(ce)
    try:
        response_post.json()['access_token']
        return response_post
    except KeyError:
        return session.post(
            f'{oauth_url}?mfa_token={totp.now()}', headers=post_headers, data=post_body
        )


def get_access_token_for_sso(auth, username, password, oauth_url, post_headers, session):
    oauth_url = APIGEE_ZONENAME_OAUTH_URL.format(zonename=auth.zonename)
    passcode_url = APIGEE_SAML_LOGIN_URL.format(zonename=auth.zonename)
    webbrowser.open(passcode_url)
    console.echo('SSO authorization page has automatically been opened in your default browser.')
    console.echo('Follow the instructions in the browser to complete this authorization request.')
    console.echo(
        f"""\nIf your browser did not automatically open, go to the following URL and sign in:\n\n{passcode_url}\n\nthen copy the Temporary Authentication Code.\n"""
    )
    try:
        passcode = click.prompt('Please enter the Temporary Authentication Code')
        post_body = f'passcode={passcode}&grant_type=password&response_type=token'
        try:
            response_post = session.post(f'{oauth_url}', headers=post_headers, data=post_body)
            response_post.json()['access_token']
            return response_post
        except KeyError:
            sys.exit('Temporary Authentication Code is invalid. Please try again.')
    except ConnectionError as ce:
        console.echo(ce)
    except KeyError:
        pass


def generate_auth_error_message(error):
    error_message = f'An exception of type {type(error).__name__} occurred. Arguments:\n{error}\nDouble check your credentials and try again.'
    if APIGEE_CLI_IS_MACHINE_USER:
        return f'{error_message} \nWARNING: APIGEE_CLI_IS_MACHINE_USER={APIGEE_CLI_IS_MACHINE_USER}'
    return error_message


def get_access_token(
    auth, retries=4, backoff_factor=0.3, status_forcelist=(500, 502, 504), session=None
):
    oauth_url = APIGEE_OAUTH_URL
    username = auth.username
    password = auth.password
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session = requests.Session()
    session.mount('https://', adapter)
    post_headers = {
        'Content-Type': 'application/x-www-form-urlencoded;charset=utf-8',
        'Accept': 'application/json;charset=utf-8',
        'Authorization': 'Basic ZWRnZWNsaTplZGdlY2xpc2VjcmV0',
    }
    if auth.token or APIGEE_CLI_IS_MACHINE_USER:
        response_post = get_access_token_for_token(
            auth, username, password, oauth_url, post_headers, session
        )
    elif auth.mfa_secret:
        response_post = get_access_token_for_mfa(
            auth, username, password, oauth_url, post_headers, session
        )
    elif auth.zonename:
        response_post = get_access_token_for_sso(
            auth, username, password, oauth_url, post_headers, session
        )
    else:
        return
    try:
        return response_post.json()['access_token']
    except KeyError as ke:
        sys.exit(generate_auth_error_message(ke))


def get_credential(section, key):
    try:
        config = configparser.ConfigParser()
        config.read(APIGEE_CLI_CREDENTIALS_FILE)
        if section in config:
            return config[section][key]
    except:
        return


def set_header(auth_obj, headers={}):
    if auth_obj.mfa_secret or auth_obj.token or auth_obj.zonename:
        access_token = ""
        make_dirs(APIGEE_CLI_DIRECTORY)
        try:
            with open(APIGEE_CLI_ACCESS_TOKEN_FILE, 'r') as f:
                access_token = f.read().strip()
        except (IOError, OSError):
            pass
        if access_token:
            decoded = jwt.decode(
                access_token,
                options={'verify_exp': False, 'verify_signature': False, 'verify_aud': False},
            )
            if (
                decoded['exp'] < int(time.time())
                or decoded['email'].lower() != auth_obj.username.lower()
            ):
                access_token = ""
        if not access_token:
            access_token = get_access_token(auth_obj)
            with open(APIGEE_CLI_ACCESS_TOKEN_FILE, 'w') as f:
                f.write(access_token)
        headers['Authorization'] = f'Bearer {access_token}'
    else:
        headers['Authorization'] = (
            'Basic '
            + base64.b64encode((f'{auth_obj.username}:{auth_obj.password}').encode()).decode()
        )
    return headers


def auth_with_prefix(auth_obj, org, name, file=None, key='name'):
    if file:
        with open(file) as f:
            attr = json.loads(f.read())[key]
        return with_prefix(auth_obj, org, attr)
    return with_prefix(auth_obj, org, name)


@click.command(
    help='Custom authorization commands. More information on the use cases for these commands are yet to be documented.',
    cls=AliasedGroup,
)
def auth():
    pass


@auth.command(name='get-access-token', help='request a fresh access token')
@common_auth_options
@common_verbose_options
@common_silent_options
def get_access_token_command(
    username, password, mfa_secret, token, zonename, org, profile, **kwargs
):
    console.echo(get_access_token(gen_auth(username, password, mfa_secret, token, zonename)))


@auth.command(help='check if user (developer) is authorized to access resource with prefix in file')
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option(
    '-f',
    '--file',
    type=click.Path(exists=True, dir_okay=False, file_okay=True, resolve_path=False),
    required=True,
)
@click.option('-k', '--key', help='name of the attribute/key to check', default='name')
def file(username, password, mfa_secret, token, zonename, org, profile, file, key, **kwargs):
    console.echo(
        auth_with_prefix(
            gen_auth(username, password, mfa_secret, token, zonename), org, None, file=file, key=key
        )
    )


@auth.command(help='check if user (developer) is authorized to access resource with prefix in name')
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option('-n', '--name', help='name of the resource to check', required=True)
def name(username, password, mfa_secret, token, zonename, org, profile, name, **kwargs):
    console.echo(
        auth_with_prefix(
            gen_auth(username, password, mfa_secret, token, zonename),
            org,
            name,
            file=None,
            key='name',
        )
    )



import importlib
import inspect
import json
import logging
import os
import re
import sys
import zipfile
from pathlib import Path

import click


def add_to_dict_if_exists(options_dict, initial_dict={}):
    for k, v in options_dict.items():
        if v:
            initial_dict[k] = v
    return initial_dict


def convert_to_set(iterable):
    if not isinstance(iterable, set):
        return set(iterable)
    return iterable


def extract_zip(source, dest):
    with zipfile.ZipFile(source, 'r') as zip_ref:
        zip_ref.extractall(dest)


def generate_path_str(*args):
    if not args:
        return
    path = None
    for arg in args:
        if not path:
            path = Path(arg)
        else:
            path /= arg
    return str(path)


def is_dir(d):
    return os.path.isdir(d)


def is_envvar_true(value):
    return value in (True, 'True', 'true', '1')


def is_file(f):
    return os.path.isfile(f)


def import_all_modules_in_directory(plugins_init_file, existing_commands):
    try:
        spec = importlib.util.spec_from_file_location('plugins_modules', plugins_init_file)
        module = importlib.util.module_from_spec(spec)
        sys.modules[spec.name] = module
        spec.loader.exec_module(module)
        import plugins_modules
        from plugins_modules import __all__ as all_plugins_modules

        for module in all_plugins_modules:
            _module = getattr(plugins_modules, module)
            if isinstance(_module, (click.core.Command, click.core.Group)):
                existing_commands.add(_module)
    except ImportError:
        logging.warning(
            f'{inspect.stack()[0][3]}; will skip loading plugin: {module}', exc_info=True
        )


def make_dirs(path):
    if not os.path.exists(path):
        try:
            os.makedirs(path)
        except FileExistsError:
            logging.warning(f'{inspect.stack()[0][3]}; will ignore FileExistsError')


def path_exists(file):
    if os.path.exists(file):
        sys.exit(f'error: {file} already exists')


def paths_exist(files):
    for file in files:
        path_exists(file)


def read_file(file, type='text'):
    with open(file, 'r') as f:
        if type == 'json':
            return json.loads(f.read())
        return f.read()


def remove_file_above_size(file, size_kb=100):
    if os.path.getsize(file) > size_kb * 1024:
        os.remove(file)


def remove_last_items_from_list(init_list, integer=0):
    if integer <= 0:
        return init_list
    return init_list[:-integer]


def resolve_target_directory(target_directory=None):
    if target_directory:
        if not os.path.exists(target_directory):
            os.makedirs(target_directory)
        return str(Path(target_directory).resolve())
    return os.getcwd()


def run_func_on_dir_files(dir, func, glob='**/*', args=(), kwargs={}):
    state = []
    for file_path in Path(resolve_target_directory(dir)).resolve().glob(glob):
        _tuple = (str(file_path),)
        result = func(*(_tuple + args), **kwargs)
        if result:
            state.append(result)
    return state


def run_func_on_iterable(iterable, func, state_op='append', args=(), kwargs={}):
    state = []
    for item in iterable:
        _tuple = (item,)
        result = func(*(_tuple + args), **kwargs)
        if result:
            getattr(state, state_op)(result)
    return state


def show_message(msg):
    print(msg)


def split_path(path, delimiter='[/\\\\]'):
    return re.split(delimiter, path)


def touch(path):
    try:
        with open(path, 'x'):
            os.utime(path, None)
    except FileNotFoundError:
        os.makedirs(os.path.split(path)[0])
    except FileExistsError:
        logging.warning(f'{inspect.stack()[0][3]}; will ignore FileExistsError')


def write_file(content, path, fs_write=True, indent=None, eof=True):
    if not fs_write:
        return
    touch(path)
    with open(path, 'w') as f:
        if isinstance(content, str):
            if eof:
                content = f'{content}\n'
            f.write(content)
        elif isinstance(content, dict) or isinstance(content, list):
            if isinstance(indent, int):
                content = f'{json.dumps(content, indent=indent)}'
            else:
                content = f'{json.dumps(content)}'
            if eof:
                f.write(f'{content}\n')
            else:
                f.write(content)


def write_zip(file, content):
    touch(file)
    with open(file, 'wb') as f:
        f.write(content)



import json
import logging
import os
from pathlib import Path

from requests.exceptions import HTTPError
from tqdm import tqdm

from apigee import APIGEE_ADMIN_API_URL, auth, console
from apigee.apiproducts.apiproducts import Apiproducts
from apigee.apis.apis import Apis
from apigee.apps.apps import Apps
from apigee.caches.caches import Caches
from apigee.developers.developers import Developers
from apigee.exceptions import InvalidApisError, NotYetImplementedError
from apigee.keyvaluemaps.keyvaluemaps import Keyvaluemaps
from apigee.permissions.permissions import Permissions
from apigee.targetservers.targetservers import Targetservers
from apigee.types import APIS, Struct, empty_snapshot
from apigee.userroles.userroles import Userroles
from apigee.utils import (extract_zip, resolve_target_directory, touch,
                          write_file)


class Backups:
    def __init__(
        self,
        auth,
        org_name,
        target_directory,
        prefix=None,
        fs_write=False,
        apis=APIS,
        environments=['test', 'prod'],
    ):
        self.auth = auth
        self.org_name = org_name
        self.target_directory = resolve_target_directory(target_directory)
        self.prefix = prefix
        self.fs_write = fs_write
        if not isinstance(apis, set):
            apis = set(apis)
        self.apis = sorted(apis)
        self.environments = environments
        self.snapshot_data = empty_snapshot()
        self.snapshot_size = 0
        self.progress_bar = None
        self.org_path = Path(self.target_directory) / self.org_name

    @property
    def apis(self):
        return self._apis

    @apis.setter
    def apis(self, value):
        if not value:
            raise InvalidApisError
        for t in value:
            if t not in APIS:
                raise InvalidApisError
        self._apis = value

    @staticmethod
    def generate_download_path(root_path, is_snapshot=False, subpaths=[]):
        path = root_path
        if is_snapshot:
            path /= 'snapshots'
        for subpath in subpaths:
            path /= subpath
        return str(path)

    @staticmethod
    def log_error(error, append_msg=''):
        error_message = (
            f'Ignoring {type(error).__name__} {error.response.status_code} error' + append_msg
        )
        logging.warning(error_message)
        console.echo(error_message)

    def __progress_callback(self, desc=''):
        if not isinstance(self.progress_bar, tqdm):
            self.progress_bar = tqdm(
                total=self.snapshot_size,
                unit='files',
                bar_format='{l_bar}{bar:32}{r_bar}{bar:-10b}',
                leave=False,
            )
        if desc:
            self.progress_bar.set_description(desc)
        self.progress_bar.update(1)

    def download_apis_snapshot(self):
        for api in Apis(self.auth, self.org_name, None).list_api_proxies(
            prefix=self.prefix, format='dict'
        ):
            self.snapshot_data.apis[api] = (
                Apis(self.auth, self.org_name, None).get_api_proxy(api).json()
            )
            write_file(
                self.snapshot_data.apis[api],
                Backups.generate_download_path(
                    self.org_path, is_snapshot=True, subpaths=['apis', f'{api}.json']
                ),
                fs_write=self.fs_write,
                indent=2,
            )

    def download_apis(self):
        for api, metadata in self.snapshot_data.apis.items():
            for revision in metadata['revision']:
                output_file = Backups.generate_download_path(
                    self.org_path, subpaths=['apis', api, revision, f'{api}.zip']
                )
                target_directory = os.path.dirname(output_file)
                try:
                    Apis(self.auth, self.org_name, None).export_api_proxy(
                        api, revision, fs_write=True, output_file=output_file
                    )
                    extract_zip(output_file, target_directory)
                    os.remove(output_file)
                except HTTPError as e:
                    Backups.log_error(e, append_msg=' for API Proxy ({api}, revision {revision})')
            self._Backups__progress_callback(desc='APIs')

    def download_keyvaluemaps_snapshot(self):
        for environment in self.environments:
            self.snapshot_data.keyvaluemaps[environment] = Keyvaluemaps(
                self.auth, self.org_name, None
            ).list_keyvaluemaps_in_an_environment(environment, prefix=self.prefix, format='dict')
            write_file(
                self.snapshot_data.keyvaluemaps[environment],
                Backups.generate_download_path(
                    self.org_path,
                    is_snapshot=True,
                    subpaths=['keyvaluemaps', environment, 'keyvaluemaps.json'],
                ),
                fs_write=self.fs_write,
                indent=2,
            )

    def download_keyvaluemaps(self):
        for environment in self.environments:
            for kvm in self.snapshot_data.keyvaluemaps[environment]:
                try:
                    write_file(
                        Keyvaluemaps(self.auth, self.org_name, kvm)
                        .get_keyvaluemap_in_an_environment(environment)
                        .text,
                        Backups.generate_download_path(
                            self.org_path, subpaths=['keyvaluemaps', environment, f'{kvm}.json']
                        ),
                        fs_write=self.fs_write,
                    )
                except HTTPError as e:
                    Backups.log_error(e, append_msg=' for KVM ({kvm})')
                self._Backups__progress_callback(desc='KeyValueMaps')

    def download_targetservers_snapshot(self):
        for environment in self.environments:
            self.snapshot_data.targetservers[environment] = Targetservers(
                self.auth, self.org_name, None
            ).list_targetservers_in_an_environment(environment, prefix=self.prefix, format='dict')
            write_file(
                self.snapshot_data.targetservers[environment],
                Backups.generate_download_path(
                    self.org_path,
                    is_snapshot=True,
                    subpaths=['targetservers', environment, 'targetservers.json'],
                ),
                fs_write=self.fs_write,
                indent=2,
            )

    def download_targetservers(self):
        for environment in self.environments:
            for targetserver in self.snapshot_data.targetservers[environment]:
                try:
                    write_file(
                        Targetservers(self.auth, self.org_name, targetserver)
                        .get_targetserver(environment)
                        .text,
                        Backups.generate_download_path(
                            self.org_path,
                            subpaths=['targetservers', environment, f'{targetserver}.json'],
                        ),
                        fs_write=self.fs_write,
                    )
                except HTTPError as e:
                    Backups.log_error(e, append_msg=' for TargetServer ({targetserver})')
                self._Backups__progress_callback(desc='TargetServers')

    def download_caches_snapshot(self):
        for environment in self.environments:
            self.snapshot_data.caches[environment] = Caches(
                self.auth, self.org_name, None
            ).list_caches_in_an_environment(environment, prefix=self.prefix, format='dict')
            write_file(
                self.snapshot_data.caches[environment],
                Backups.generate_download_path(
                    self.org_path, is_snapshot=True, subpaths=['caches', environment, 'caches.json']
                ),
                fs_write=self.fs_write,
                indent=2,
            )

    def download_caches(self):
        for environment in self.environments:
            for cache in self.snapshot_data.caches[environment]:
                try:
                    write_file(
                        Caches(self.auth, self.org_name, cache)
                        .get_information_about_a_cache(environment)
                        .text,
                        Backups.generate_download_path(
                            self.org_path, subpaths=['caches', environment, f'{cache}.json']
                        ),
                        fs_write=self.fs_write,
                    )
                except HTTPError as e:
                    Backups.log_error(e, append_msg=' for Cache ({cache})')
                self._Backups__progress_callback(desc='Caches')

    def download_developers_snapshot(self):
        self.snapshot_data.developers = Developers(self.auth, self.org_name, None).list_developers(
            prefix=self.prefix, format='dict'
        )
        write_file(
            self.snapshot_data.developers,
            Backups.generate_download_path(
                self.org_path, is_snapshot=True, subpaths=['developers', 'developers.json']
            ),
            fs_write=self.fs_write,
            indent=2,
        )

    def download_developers(self):
        for developer in self.snapshot_data.developers:
            try:
                write_file(
                    Developers(self.auth, self.org_name, developer).get_developer().text,
                    Backups.generate_download_path(
                        self.org_path, subpaths=['developers', f'{developer}.json']
                    ),
                    fs_write=self.fs_write,
                )
            except HTTPError as e:
                Backups.log_error(e, append_msg=' for Developer ({developer})')
            self._Backups__progress_callback(desc='Developers')

    def download_apiproducts_snapshot(self):
        self.snapshot_data.apiproducts = Apiproducts(
            self.auth, self.org_name, None
        ).list_api_products(prefix=self.prefix, format='dict')
        write_file(
            self.snapshot_data.apiproducts,
            Backups.generate_download_path(
                self.org_path, is_snapshot=True, subpaths=['apiproducts', 'apiproducts.json']
            ),
            fs_write=self.fs_write,
            indent=2,
        )

    def download_apiproducts(self):
        for apiproduct in self.snapshot_data.apiproducts:
            try:
                write_file(
                    Apiproducts(self.auth, self.org_name, apiproduct).get_api_product().text,
                    Backups.generate_download_path(
                        self.org_path, subpaths=['apiproducts', f'{apiproduct}.json']
                    ),
                    fs_write=self.fs_write,
                )
            except HTTPError as e:
                Backups.log_error(e, append_msg=' for API Product ({apiproduct})')
            self._Backups__progress_callback(desc='API Products')

    def download_apps_snapshot(self, expand=False, count=1000, startkey=""):
        self.snapshot_data.apps = Apps(self.auth, self.org_name, None).list_apps_for_all_developers(
            Developers(self.auth, self.org_name, None).list_developers(
                prefix=None, expand=expand, count=count, startkey=startkey, format='dict'
            ),
            prefix=self.prefix,
            format='dict',
        )
        self.snapshot_data.apps = {k: v for k, v in self.snapshot_data.apps.items() if v}
        for app, details in self.snapshot_data.apps.items():
            write_file(
                details,
                Backups.generate_download_path(
                    self.org_path, is_snapshot=True, subpaths=['apps', f'{app}.json']
                ),
                fs_write=self.fs_write,
                indent=2,
            )

    def download_apps(self):
        for developer, apps in self.snapshot_data.apps.items():
            for app in apps:
                try:
                    write_file(
                        Apps(self.auth, self.org_name, app)
                        .get_developer_app_details(developer)
                        .text,
                        Backups.generate_download_path(
                            self.org_path, subpaths=['apps', developer, f'{app}.json']
                        ),
                        fs_write=self.fs_write,
                    )
                except HTTPError as e:
                    Backups.log_error(e, append_msg=' for Developer App ({app})')
                self._Backups__progress_callback(desc='Developer Apps')

    def download_userroles_snapshot(self):
        self.snapshot_data.userroles = (
            Userroles(self.auth, self.org_name, None).list_user_roles().json()
        )
        if self.prefix:
            self.snapshot_data.userroles = [
                role for role in self.snapshot_data.userroles if role.startswith(self.prefix)
            ]
        write_file(
            self.snapshot_data.userroles,
            Backups.generate_download_path(
                self.org_path, is_snapshot=True, subpaths=['userroles', 'userroles.json']
            ),
            fs_write=self.fs_write,
            indent=2,
        )

    def download_users_for_a_role(self, role_name):
        try:
            write_file(
                Userroles(self.auth, self.org_name, role_name).get_users_for_a_role().json(),
                Backups.generate_download_path(
                    self.org_path, subpaths=['userroles', role_name, 'users.json']
                ),
                fs_write=self.fs_write,
                indent=2,
            )
        except HTTPError as e:
            Backups.log_error(e, append_msg=' for User Role ({role_name}) users')

    def download_resource_permissions(self, role_name):
        try:
            write_file(
                json.dumps(
                    Userroles.sort_permissions(
                        Permissions(self.auth, self.org_name, role_name).get_permissions(
                            formatted=True, format='json'
                        )
                    ),
                    indent=2,
                ),
                Backups.generate_download_path(
                    self.org_path, subpaths=['userroles', role_name, 'resource_permissions.json']
                ),
                fs_write=self.fs_write,
            )
        except HTTPError as e:
            Backups.log_error(e, append_msg=' for User Role ({role_name}) resource permissions')

    def download_userroles(self):
        for role_name in self.snapshot_data.userroles:
            self.download_users_for_a_role(role_name)
            self.download_resource_permissions(role_name)
            self._Backups__progress_callback(desc='User Roles')

    def __calculate_snapshot_size(self):
        count = 0
        for x in self.snapshot_data.__dict__:
            if x == 'apis':
                count += len(self.snapshot_data.apis)
            elif x in ('keyvaluemaps', 'targetservers', 'caches'):
                for environment_bound_api, listing in self.snapshot_data.__dict__[x].items():
                    count += len(listing)
            elif x == 'apps':
                for developer, apps in self.snapshot_data.apps.items():
                    count += len(apps)
            elif isinstance(self.snapshot_data.__dict__[x], list):
                count += len(self.snapshot_data.__dict__[x])
        return count

    def get_snapshots(self):
        for api in self.apis:
            if api in {'apis', 'apps'}:
                console.echo(
                    f'Retrieving {api} listing (this may take a while)... ', end='', flush=True
                )
            else:
                console.echo(f'Retrieving {api} listing... ', end='', flush=True)
            getattr(self, f'download_{api}_snapshot')()
            console.echo('Done')
        self.snapshot_size = self._Backups__calculate_snapshot_size()

    def take_snapshot(self):
        self.get_snapshots()
        console.echo('Generating snapshot files...')
        for api in self.apis:
            getattr(self, f'download_{api}')()
        self.progress_bar.close()
        console.echo('Done.')



import json


class InformalApisInterface:
    def __init__(self, auth, org_name):
        self._auth = auth
        self._org_name = org_name

    def __call__(self):
        pass

    @property
    def auth(self):
        return self._auth

    @auth.setter
    def auth(self, value):
        self._auth = value

    @property
    def org_name(self):
        return self._org_name

    @org_name.setter
    def org_name(self, value):
        self._org_name = value



import builtins

import click


def silent_callback(ctx, param, value):
    builtins.APIGEE_CLI_TOGGLE_SILENT = value


def common_silent_options(func):
    func = click.option(
        '--silent',
        show_default='toggle silent output',
        flag_value=True,
        default=False,
        callback=silent_callback,
    )(func)
    return func



import json
import random
import string

import requests
from requests.exceptions import HTTPError

from apigee import (APIGEE_ADMIN_API_URL,
                    APIGEE_CLI_AUTHORIZATION_DEVELOPER_ATTRIBUTE, auth,
                    console)
from apigee.developers.serializer import DevelopersSerializer

CREATE_DEVELOPER_PATH = '{api_url}/v1/organizations/{org}/developers'
DELETE_DEVELOPER_PATH = '{api_url}/v1/organizations/{org}/developers/{developer_email}'
GET_DEVELOPER_PATH = '{api_url}/v1/organizations/{org}/developers/{developer_email}'
GET_DEVELOPER_BY_APP_PATH = '{api_url}/v1/organizations/{org}/developers?app={app_name}'
LIST_DEVELOPERS_PATH = (
    '{api_url}/v1/organizations/{org}/developers?expand={expand}&count={count}&startKey={startkey}'
)
SET_DEVELOPER_STATUS_PATH = (
    '{api_url}/v1/organizations/{org}/developers/{developer_email}?action={action}'
)
UPDATE_DEVELOPER_PATH = '{api_url}/v1/organizations/{org}/developers/{developer_email}'
GET_DEVELOPER_ATTRIBUTE_PATH = (
    '{api_url}/v1/organizations/{org}/developers/{developer_email}/attributes/{attribute_name}'
)
UPDATE_A_DEVELOPER_ATTRIBUTE_PATH = (
    '{api_url}/v1/organizations/{org}/developers/{developer_email}/attributes/{attribute_name}'
)
DELETE_DEVELOPER_ATTRIBUTE_PATH = (
    '{api_url}/v1/organizations/{org}/developers/{developer_email}/attributes/{attribute_name}'
)
GET_ALL_DEVELOPER_ATTRIBUTES_PATH = (
    '{api_url}/v1/organizations/{org}/developers/{developer_email}/attributes'
)
UPDATE_ALL_DEVELOPER_ATTRIBUTES_PATH = (
    '{api_url}/v1/organizations/{org}/developers/{developer_email}/attributes'
)


class Developers:
    def __init__(self, auth, org_name, developer_email):
        self._auth = auth
        self._org_name = org_name
        self._developer_email = developer_email

    @property
    def auth(self):
        return self._auth

    @auth.setter
    def auth(self, value):
        self._auth = value

    @property
    def org_name(self):
        return self._org_name

    @org_name.setter
    def org_name(self, value):
        self._org_name = value

    @property
    def developer_email(self):
        return self._developer_email

    @developer_email.setter
    def developer_email(self, value):
        self._developer_email = value

    def __call__(self):
        pass

    def create_developer(self, first_name, last_name, user_name, attributes='{"attributes" : [ ]}'):
        uri = CREATE_DEVELOPER_PATH.format(api_url=APIGEE_ADMIN_API_URL, org=self._org_name)
        hdrs = auth.set_header(
            self._auth, headers={'Accept': 'application/json', 'Content-Type': 'application/json'}
        )
        body = {
            'email': self._developer_email,
            'firstName': first_name,
            'lastName': last_name,
            'userName': user_name,
            'attributes': json.loads(attributes)['attributes'],
        }
        resp = requests.post(uri, headers=hdrs, json=body)
        resp.raise_for_status()
        return resp

    def delete_developer(self):
        uri = DELETE_DEVELOPER_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, developer_email=self._developer_email
        )
        hdrs = auth.set_header(self._auth, headers={'Accept': 'application/json'})
        resp = requests.delete(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def get_developer(self):
        uri = GET_DEVELOPER_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, developer_email=self._developer_email
        )
        hdrs = auth.set_header(self._auth, headers={'Accept': 'application/json'})
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def get_developer_by_app(self, app_name):
        uri = GET_DEVELOPER_BY_APP_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, app_name=app_name
        )
        hdrs = auth.set_header(self._auth, headers={'Accept': 'application/json'})
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def list_developers(self, prefix=None, expand=False, count=1000, startkey="", format='json'):
        uri = LIST_DEVELOPERS_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            expand=expand,
            count=count,
            startkey=startkey,
        )
        hdrs = auth.set_header(self._auth, headers={'Accept': 'application/json'})
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return DevelopersSerializer().serialize_details(resp, format, prefix=prefix)

    def set_developer_status(self, action):
        uri = SET_DEVELOPER_STATUS_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            developer_email=self._developer_email,
            action=action,
        )
        hdrs = auth.set_header(
            self._auth,
            headers={'Accept': 'application/json', 'Content-Type': 'application/octet-stream'},
        )
        resp = requests.post(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def update_developer(self, request_body):
        uri = UPDATE_DEVELOPER_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, developer_email=self._developer_email
        )
        hdrs = auth.set_header(
            self._auth, headers={'Accept': 'application/json', 'Content-Type': 'application/json'}
        )
        body = json.loads(request_body)
        resp = requests.put(uri, headers=hdrs, json=body)
        resp.raise_for_status()
        return resp

    def get_developer_attribute(self, attribute_name):
        uri = GET_DEVELOPER_ATTRIBUTE_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            developer_email=self._developer_email,
            attribute_name=attribute_name,
        )
        hdrs = auth.set_header(self._auth, headers={'Accept': 'application/json'})
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def update_a_developer_attribute(self, attribute_name, updated_value):
        uri = UPDATE_A_DEVELOPER_ATTRIBUTE_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            developer_email=self._developer_email,
            attribute_name=attribute_name,
        )
        hdrs = auth.set_header(self._auth, headers={'Accept': 'application/json'})
        body = {'value': updated_value}
        resp = requests.post(uri, headers=hdrs, json=body)
        resp.raise_for_status()
        return resp

    def delete_developer_attribute(self, attribute_name):
        uri = DELETE_DEVELOPER_ATTRIBUTE_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            developer_email=self._developer_email,
            attribute_name=attribute_name,
        )
        hdrs = auth.set_header(self._auth, headers={'Accept': 'application/json'})
        resp = requests.delete(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def get_all_developer_attributes(self):
        uri = GET_ALL_DEVELOPER_ATTRIBUTES_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, developer_email=self._developer_email
        )
        hdrs = auth.set_header(self._auth, headers={'Accept': 'application/json'})
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def update_all_developer_attributes(self, request_body):
        uri = UPDATE_ALL_DEVELOPER_ATTRIBUTES_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, developer_email=self._developer_email
        )
        hdrs = auth.set_header(
            self._auth, headers={'Accept': 'application/json', 'Content-Type': 'application/json'}
        )
        body = json.loads(request_body)
        resp = requests.post(uri, headers=hdrs, json=body)
        resp.raise_for_status()
        return resp



import json

from tabulate import tabulate


class PermissionsSerializer:
    def serialize_details(self, permission_details, format, showindex=False, tablefmt='plain'):
        if format == 'text':
            return permission_details.text
        elif format == 'json':
            return permission_details.json()
        elif format == 'table':
            table = [
                [res['organization'], res['path'], res['permissions']]
                for res in permission_details.json()['resourcePermission']
            ]
            headers = []
            if showindex == 'always' or showindex is True:
                headers = ['id', 'organization', 'path', 'permissions']
            elif showindex == 'never' or showindex is False:
                headers = ['organization', 'path', 'permissions']
            return tabulate(table, headers, showindex=showindex, tablefmt=tablefmt)
        return permission_details



import builtins
import json
from os import getenv
from pathlib import Path

from apigee import utils

APP = 'apigeecli'
CMD = 'apigee'
__version__ = '0.50.1'
description = '(Unofficial) Apigee Management API command-line interface with multi-factor authentication (MFA) and single sign-on (SSO)/SAML support'
long_description = """(Unofficial) Apigee Management API command-line interface with multi-factor authentication (MFA) and single sign-on (SSO)/SAML support"""

APIGEE_CLI_DIRECTORY = utils.generate_path_str(Path.home(), '.apigee')
APIGEE_CLI_PLUGINS_DIRECTORY = utils.generate_path_str(APIGEE_CLI_DIRECTORY, 'plugins')

APIGEE_ADMIN_API_URL = 'https://api.enterprise.apigee.com'
APIGEE_CLI_ACCESS_TOKEN_FILE = utils.generate_path_str(APIGEE_CLI_DIRECTORY, 'access_token')
APIGEE_CLI_AUTHORIZATION_DEVELOPER_ATTRIBUTE = 'team'
APIGEE_CLI_CREDENTIALS_FILE = utils.generate_path_str(APIGEE_CLI_DIRECTORY, 'credentials')
# APIGEE_CLI_EXCEPTION_LOG_FILE = utils.generate_path_str(APIGEE_CLI_DIRECTORY, 'exception.log')
APIGEE_CLI_EXCEPTIONS_LOG_FILE = utils.generate_path_str(APIGEE_CLI_DIRECTORY, 'exceptions.log')
APIGEE_CLI_IS_MACHINE_USER = utils.is_envvar_true(getenv('APIGEE_CLI_IS_MACHINE_USER'))
APIGEE_CLI_PLUGINS_CONFIG_FILE = utils.generate_path_str(APIGEE_CLI_PLUGINS_DIRECTORY, 'config')
APIGEE_CLI_PLUGINS_PATH = utils.generate_path_str(APIGEE_CLI_PLUGINS_DIRECTORY, '__init__.py')
APIGEE_CLI_PREFIX = getenv('APIGEE_CLI_PREFIX')
APIGEE_CLI_SYMMETRIC_KEY = getenv('APIGEE_CLI_SYMMETRIC_KEY')
APIGEE_CLI_TOGGLE_SILENT = False
APIGEE_CLI_TOGGLE_VERBOSE = 0
APIGEE_IS_TOKEN = getenv('APIGEE_IS_TOKEN')
APIGEE_MFA_SECRET = getenv('APIGEE_MFA_SECRET')
APIGEE_OAUTH_URL = 'https://login.apigee.com/oauth/token'
APIGEE_ORG = getenv('APIGEE_ORG')
APIGEE_PASSWORD = getenv('APIGEE_PASSWORD')
APIGEE_SAML_LOGIN_URL = 'https://{zonename}.login.apigee.com/passcode'
APIGEE_USERNAME = getenv('APIGEE_USERNAME')
APIGEE_ZONENAME = getenv('APIGEE_ZONENAME')
APIGEE_ZONENAME_OAUTH_URL = 'https://{zonename}.login.apigee.com/oauth/token'

builtins.APIGEE_CLI_TOGGLE_SILENT = APIGEE_CLI_TOGGLE_SILENT
builtins.APIGEE_CLI_TOGGLE_VERBOSE = APIGEE_CLI_TOGGLE_VERBOSE



#!/usr/bin/env python

# Copyright 2019 Matthew Delotavo
# Copyright 2015 Apigee Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import getopt
import http.client
import io
import json
import os
import re
import sys
import urllib.parse
import xml.dom.minidom
import zipfile

from requests.exceptions import HTTPError

from apigee import APIGEE_ADMIN_API_URL, auth, console
from apigee.apis.apis import Apis

ApigeeHost = APIGEE_ADMIN_API_URL
UserPW = None
Directory = None
Organization = None
Environment = None
Name = None
BasePath = '/'
ShouldDeploy = True
ShouldOverride = False
GracePeriod = 15
Auth = None

url = urllib.parse.urlparse(ApigeeHost)
httpScheme = url[0]
httpHost = url[1]

body = None


class Struct:
    def __init__(self, **entries):
        self.__dict__.update(entries)


def httpCall(verb, uri, headers, body):
    if httpScheme == 'https':
        conn = http.client.HTTPSConnection(httpHost)
    else:
        conn = http.client.HTTPConnection(httpHost)

    if headers == None:
        hdrs = {}
    else:
        hdrs = headers

    # if Auth:
    #     # hdrs['Authorization'] = 'Bearer %s' % AccessToken
    #     hdrs = auth.set_header(hdrs, Auth)
    # else:
    #     hdrs['Authorization'] = 'Basic %s' % base64.b64encode(UserPW.encode()).decode()
    hdrs = auth.set_header(Auth, headers=hdrs)
    conn.request(verb, uri, body, hdrs)

    return conn.getresponse()


def getElementText(n):
    c = n.firstChild
    str = io.StringIO()

    while c != None:
        if c.nodeType == xml.dom.Node.TEXT_NODE:
            str.write(c.data)
        c = c.nextSibling

    return str.getvalue().strip()


def getElementVal(n, name):
    c = n.firstChild

    while c != None:
        if c.nodeName == name:
            return getElementText(c)
        c = c.nextSibling

    return None


# Return TRUE if any component of the file path contains a directory name that
# starts with a "." like '.svn', but not '.' or '..'
def pathContainsDot(p):
    c = re.compile('\.\w+')

    for pc in p.split('/'):
        if c.match(pc) != None:
            return True

    return False


def getDeployments():
    # Print info on deployments
    hdrs = {'Accept': 'application/xml'}
    resp = httpCall(
        'GET', '/v1/organizations/%s/apis/%s/deployments' % (Organization, Name), hdrs, None
    )

    if resp.status != 200:
        return None

    ret = []
    deployments = xml.dom.minidom.parse(resp)
    environments = deployments.getElementsByTagName('Environment')

    for env in environments:
        envName = env.getAttribute('name')
        revisions = env.getElementsByTagName('Revision')
        for rev in revisions:
            revNum = int(rev.getAttribute('name'))
            error = None
            state = getElementVal(rev, 'State')
            basePaths = rev.getElementsByTagName('BasePath')

            if len(basePaths) > 0:
                basePath = getElementText(basePaths[0])
            else:
                basePath = 'unknown'

            # svrs = rev.getElementsByTagName('Server')
            status = {
                'environment': envName,
                'revision': revNum,
                'basePath': basePath,
                'state': state,
            }

            if error != None:
                status['error'] = error

            ret.append(status)

    return ret


def printDeployments(dep, check_revision=None):
    if check_revision:
        revisions = [d['revision'] for d in dep]
        if check_revision not in revisions:
            sys.exit('Error: proxy version %i not found' % check_revision)
        console.echo('Proxy version %i found' % check_revision)
    for d in dep:
        console.echo('Environment: %s' % d['environment'])
        console.echo('  Revision: %i BasePath = %s' % (d['revision'], d['basePath']))
        console.echo('  State: %s' % d['state'])
        if d['state'] == 'missing':
            console.echo('Missing deployment. Attempting deletion...')
            try:
                Apis(Auth, Organization).undeploy_api_proxy_revision(
                    Name, d['environment'], d['revision']
                )
                console.echo(
                    Apis(Auth, Organization).delete_api_proxy_revision(Name, d['revision']).text
                )
            except HTTPError as e:
                if e.response.status_code != 400:
                    raise e
        elif d['state'] != 'deployed':
            sys.exit(1)
        if 'error' in d:
            console.echo('  Error: %s' % d['error'])


def deploy(args):
    global UserPW
    global Directory
    global Organization
    global Environment
    global Name
    global ShouldDeploy
    global ShouldOverride
    global Auth

    # ApigeeHost = 'https://api.enterprise.apigee.com'
    UserPW = args.username + ':' + args.password
    Directory = args.directory
    Organization = args.org
    Environment = args.environment
    Name = args.name
    # BasePath = '/'
    ShouldDeploy = not args.import_only
    ShouldOverride = args.seamless_deploy
    # GracePeriod = 15
    # AccessToken = mfa_with_pyotp.get_access_token(args)
    Auth = Struct(
        username=args.username,
        password=args.password,
        mfa_secret=args.mfa_secret,
        token=args.token,
        zonename=args.zonename,
    )

    # if UserPW == None or \
    #         (Directory == None and ZipFile == None) or \
    #         Environment == None or \
    #         Name == None or \
    #         Organization == None:
    #     print """Usage: deploy -n [name] (-d [directory name] | -z [zipfile])
    #               -e [environment] -u [username:password] -o [organization]
    #               [-p [base path] -h [apigee API url] -i]
    #     base path defaults to "/"
    #     Apigee URL defaults to "https://api.enterprise.apigee.com"
    #     -i denotes to import only and not actually deploy
    #     """
    #     sys.exit(1)

    # url = urlparse.urlparse(ApigeeHost)
    # httpScheme = url[0]
    # httpHost = url[1]
    #
    # body = None

    if Directory != None:
        # Construct a ZIPped copy of the bundle in memory
        tf = io.BytesIO()
        zipout = zipfile.ZipFile(tf, 'w')

        dirList = os.walk(Directory)
        for dirEntry in dirList:
            if not pathContainsDot(dirEntry[0]):
                for fileEntry in dirEntry[2]:
                    if not fileEntry.endswith('~'):
                        fn = os.path.join(dirEntry[0], fileEntry)
                        en = os.path.join(os.path.relpath(dirEntry[0], Directory), fileEntry)
                        console.echo('Writing %s to %s' % (fn, en))
                        zipout.write(fn, en)

        zipout.close()
        body = tf.getvalue()
    elif ZipFile != None:
        f = open(ZipFile, 'r')
        body = f.read()
        f.close()

    # Upload the bundle to the API
    hdrs = {'Content-Type': 'application/octet-stream', 'Accept': 'application/json'}
    uri = '/v1/organizations/%s/apis?action=import&name=%s' % (Organization, Name)
    resp = httpCall('POST', uri, hdrs, body)

    if resp.status != 200 and resp.status != 201:
        console.echo(
            'Import failed to %s with status %i:\n%s' % (uri, resp.status, resp.read().decode())
        )
        sys.exit(2)

    deployment = json.loads(resp.read().decode())
    revision = int(deployment['revision'])

    console.echo('Imported new proxy version %i' % revision)

    if ShouldDeploy and not ShouldOverride:
        # Undeploy duplicates
        deps = getDeployments()
        for d in deps:
            if (
                d['environment'] == Environment
                and d['basePath'] == BasePath
                and d['revision'] != revision
            ):
                console.echo(
                    'Undeploying revision %i in same environment and path:' % d['revision']
                )
                conn = http.client.HTTPSConnection(httpHost)
                resp = httpCall(
                    'POST',
                    (
                        '/v1/organizations/%s/apis/%s/deployments'
                        + '?action=undeploy'
                        + '&env=%s'
                        + '&revision=%i'
                    )
                    % (Organization, Name, Environment, d['revision']),
                    None,
                    None,
                )
                if resp.status != 200 and resp.status != 204:
                    console.echo(
                        'Error %i on undeployment:\n%s' % (resp.status, resp.read().decode())
                    )

        # Deploy the bundle
        hdrs = {'Accept': 'application/json'}
        resp = httpCall(
            'POST',
            (
                '/v1/organizations/%s/apis/%s/deployments'
                + '?action=deploy'
                + '&env=%s'
                + '&revision=%i'
                + '&basepath=%s'
            )
            % (Organization, Name, Environment, revision, BasePath),
            hdrs,
            None,
        )

        if resp.status != 200 and resp.status != 201:
            console.echo('Deploy failed with status %i:\n%s' % (resp.status, resp.read().decode()))
            sys.exit(2)

    if ShouldOverride:
        # Seamless Deploy the bundle
        console.echo('Seamless deploy %s' % Name)
        hdrs = {'Content-Type': 'application/x-www-form-urlencoded'}
        resp = httpCall(
            'POST',
            (
                '/v1/organizations/%s/environments/%s/apis/%s/revisions/%s/deployments'
                + '?override=true'
                + '&delay=%s'
            )
            % (Organization, Environment, Name, revision, GracePeriod),
            hdrs,
            None,
        )

        if resp.status != 200 and resp.status != 201:
            console.echo('Deploy failed with status %i:\n%s' % (resp.status, resp.read().decode()))
            sys.exit(2)

    deps = getDeployments()
    # printDeployments(deps, check_revision=revision)
    if ShouldDeploy and not ShouldOverride:
        printDeployments(deps)
    if ShouldOverride:
        printDeployments(deps, check_revision=revision)


def main():
    pass


if __name__ == '__main__':
    main()



import json


class DevelopersSerializer:
    def serialize_details(self, developers, format, prefix=None):
        resp = developers
        if format == 'text':
            return developers.text
        developers = developers.json()
        if prefix:
            developers = [developer for developer in developers if developer.startswith(prefix)]
        if format == 'json':
            return json.dumps(developers)
        elif format == 'table':
            pass
        elif format == 'dict':
            return developers
        return resp



import click

from apigee import console
from apigee.auth import common_auth_options, gen_auth
from apigee.prefix import common_prefix_options
from apigee.silent import common_silent_options
from apigee.verbose import common_verbose_options
from apigee.virtualhosts.virtualhosts import Virtualhosts


@click.group(
    help="A named network configuration (including URL) for an environment (for example 'test' or 'prod') on API Services."
)
def virtualhosts():
    pass


def _create_a_virtual_host_for_an_environment(*args, **kwargs):
    pass


@virtualhosts.command(help='')
@common_auth_options
@common_silent_options
@common_verbose_options
def create(*args, **kwargs):
    _create_a_virtual_host_for_an_environment(*args, **kwargs)


def _delete_a_virtual_host_from_an_environment(*args, **kwargs):
    pass


@virtualhosts.command(help='')
@common_auth_options
@common_silent_options
@common_verbose_options
def delete(*args, **kwargs):
    _delete_a_virtual_host_from_an_environment(*args, **kwargs)


def _get_a_virtual_host_for_an_environment(
    username, password, mfa_secret, token, zonename, org, profile, name, environment, **kwargs
):
    return (
        Virtualhosts(gen_auth(username, password, mfa_secret, token, zonename), org, name)
        .get_a_virtual_host_for_an_environment(environment)
        .text
    )


@virtualhosts.command(help='Gets details for a named virtual host.')
@common_auth_options
@common_silent_options
@common_verbose_options
@click.option('-n', '--name', help='reference name', required=True)
@click.option('-e', '--environment', help='environment', required=True)
def get(*args, **kwargs):
    console.echo(_get_a_virtual_host_for_an_environment(*args, **kwargs))


def _list_virtual_hosts_for_an_environment(
    username,
    password,
    mfa_secret,
    token,
    zonename,
    org,
    profile,
    environment,
    prefix=None,
    **kwargs
):
    return Virtualhosts(
        gen_auth(username, password, mfa_secret, token, zonename), org, None
    ).list_virtual_hosts_for_an_environment(environment, prefix=prefix)


@virtualhosts.command(help='Get a list of named virtual hosts for an environment.')
@common_auth_options
@common_prefix_options
@common_silent_options
@common_verbose_options
@click.option('-e', '--environment', help='environment', required=True)
def list(*args, **kwargs):
    console.echo(_list_virtual_hosts_for_an_environment(*args, **kwargs))


def _update_virtual_host_for_an_environment(*args, **kwargs):
    pass


@virtualhosts.command(help='')
@common_auth_options
@common_silent_options
@common_verbose_options
def update(*args, **kwargs):
    _update_virtual_host_for_an_environment(*args, **kwargs)



import json

import requests

from apigee import APIGEE_ADMIN_API_URL, auth

ADD_A_USER_TO_A_ROLE_PATH = (
    '{api_url}/v1/organizations/{org}/userroles/{role_name}/users?id={user_email}'
)
ADD_PERMISSIONS_FOR_A_RESOURCE_TO_A_USER_ROLE_PATH = (
    '{api_url}/v1/organizations/{org}/userroles/{role_name}/permissions'
)
ADD_PERMISSIONS_FOR_MULTIPLE_RESOURCES_TO_A_USER_ROLE_PATH = (
    '{api_url}/v1/organizations/{org}/userroles/{role_name}/resourcepermissions'
)
CREATE_A_USER_ROLE_IN_AN_ORGANIZATION_PATH = '{api_url}/v1/organizations/{org}/userroles'
DELETE_A_PERMISSION_FOR_A_RESOURCE_PATH = '{api_url}/v1/organizations/{org}/userroles/{role_name}/permissions/{permission}?path={resource_path}'
DELETE_RESOURCE_FROM_PERMISSIONS_PATH = '{api_url}/v1/organizations/{org}/userroles/{role_name}/permissions?path={resource_path}&delete=true'
DELETE_A_USER_ROLE_PATH = '{api_url}/v1/organizations/{org}/userroles/{role_name}'
GET_A_ROLE_PATH = '{api_url}/v1/organizations/{org}/userroles/{role_name}'
GET_RESOURCE_PERMISSIONS_FOR_A_SPECIFIC_ROLE_PATH = (
    '{api_url}/v1/organizations/{org}/userroles/{role_name}/permissions'
)
GET_USERS_FOR_A_ROLE_PATH = '{api_url}/v1/organizations/{org}/userroles/{role_name}/users'
LIST_PERMISSIONS_FOR_A_RESOURCE_PATH = (
    '{api_url}/v1/organizations/{org}/userroles/{role_name}/permissions'
)
LIST_USER_ROLES_PATH = '{api_url}/v1/organizations/{org}/userroles'
REMOVE_USER_MEMBERSHIP_IN_ROLE_PATH = (
    '{api_url}/v1/organizations/{org}/userroles/{role_name}/users/{user_email}'
)
VERIFY_A_USER_ROLES_PERMISSION_ON_A_SPECIFIC_RBAC_RESOURCE_PATH = '{api_url}/v1/organizations/{org}/userroles/{role_name}/permissions/{permission}?path={resource_path}'
VERIFY_USER_ROLE_MEMBERSHIP_PATH = (
    '{api_url}/v1/organizations/{org}/userroles/{role_name}/users/{user_email}'
)


class Userroles:
    def __init__(self, auth, org_name, role_name):
        self._auth = auth
        self._org_name = org_name
        self._role_name = role_name

    def __call__(self):
        pass

    @property
    def auth(self):
        return self._auth

    @auth.setter
    def auth(self, value):
        self._auth = value

    @property
    def org_name(self):
        return self._org_name

    @org_name.setter
    def org_name(self, value):
        self._org_name = value

    @property
    def role_name(self):
        return self._role_name

    @role_name.setter
    def role_name(self, value):
        self._role_name = value

    def add_a_user_to_a_role(self, user_email):
        uri = ADD_A_USER_TO_A_ROLE_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            role_name=self._role_name,
            user_email=user_email,
        )
        hdrs = auth.set_header(
            self._auth,
            headers={
                'Accept': 'application/json',
                'Content-Type': 'application/x-www-form-urlencoded',
            },
        )
        resp = requests.post(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def add_permissions_for_a_resource_to_a_user_role(self, request_body):
        uri = ADD_PERMISSIONS_FOR_A_RESOURCE_TO_A_USER_ROLE_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, role_name=self._role_name
        )
        hdrs = auth.set_header(
            self._auth, headers={'Accept': 'application/json', 'Content-Type': 'application/json'}
        )
        body = json.loads(request_body)
        resp = requests.post(uri, headers=hdrs, json=body)
        resp.raise_for_status()
        return resp

    def add_permissions_for_multiple_resources_to_a_user_role(self, request_body):
        uri = f'{APIGEE_ADMIN_API_URL}/v1/organizations/{self._org_name}/userroles/{self._role_name}/resourcepermissions'
        uri = ADD_PERMISSIONS_FOR_MULTIPLE_RESOURCES_TO_A_USER_ROLE_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, role_name=self._role_name
        )
        hdrs = auth.set_header(
            self._auth, headers={'Accept': 'application/json', 'Content-Type': 'application/json'}
        )
        body = json.loads(request_body)
        resp = requests.post(uri, headers=hdrs, json=body)
        resp.raise_for_status()
        return resp

    def create_a_user_role_in_an_organization(self, roles):
        uri = CREATE_A_USER_ROLE_IN_AN_ORGANIZATION_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name
        )
        hdrs = auth.set_header(
            self._auth, headers={'Accept': 'application/json', 'Content-Type': 'application/json'}
        )
        body = {'role': [{'name': role} for role in roles]}
        resp = requests.post(uri, headers=hdrs, json=body)
        resp.raise_for_status()
        return resp

    def delete_a_permission_for_a_resource(self, permission, resource_path):
        uri = DELETE_A_PERMISSION_FOR_A_RESOURCE_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            role_name=self._role_name,
            permission=permission,
            resource_path=resource_path,
        )
        hdrs = auth.set_header(
            self._auth,
            headers={'Accept': 'application/json', 'Content-Type': 'application/octet-stream'},
        )
        resp = requests.delete(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def delete_resource_from_permissions(self, resource_path):
        uri = DELETE_RESOURCE_FROM_PERMISSIONS_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            role_name=self._role_name,
            resource_path=resource_path,
        )
        hdrs = auth.set_header(
            self._auth,
            headers={'Accept': 'application/json', 'Content-Type': 'application/octet-stream'},
        )
        resp = requests.delete(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def delete_a_user_role(self):
        uri = DELETE_A_USER_ROLE_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, role_name=self._role_name
        )
        hdrs = auth.set_header(
            self._auth,
            headers={
                'Accept': 'application/json',
                'Content-Type': 'application/x-www-form-urlencoded',
            },
        )
        resp = requests.delete(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def get_a_role(self):
        uri = GET_A_ROLE_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, role_name=self._role_name
        )
        hdrs = auth.set_header(
            self._auth, headers={'Accept': 'application/json', 'Content-Type': 'application/json'}
        )
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def get_resource_permissions_for_a_specific_role(self, resource_path=""):
        uri = GET_RESOURCE_PERMISSIONS_FOR_A_SPECIFIC_ROLE_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, role_name=self._role_name
        )
        if resource_path:
            uri += f'?path={resource_path}'
        hdrs = auth.set_header(
            self._auth,
            headers={'Accept': 'application/json', 'Content-Type': 'application/octet-stream'},
        )
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def get_users_for_a_role(self):
        uri = GET_USERS_FOR_A_ROLE_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, role_name=self._role_name
        )
        hdrs = auth.set_header(
            self._auth, headers={'Accept': 'application/json', 'Content-Type': 'application/json'}
        )
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def list_permissions_for_a_resource(self):
        uri = LIST_PERMISSIONS_FOR_A_RESOURCE_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, role_name=self._role_name
        )
        hdrs = auth.set_header(
            self._auth, headers={'Accept': 'application/json', 'Content-Type': 'application/json'}
        )
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def list_user_roles(self):
        uri = LIST_USER_ROLES_PATH.format(api_url=APIGEE_ADMIN_API_URL, org=self._org_name)
        hdrs = auth.set_header(
            self._auth, headers={'Accept': 'application/json', 'Content-Type': 'application/json'}
        )
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def remove_user_membership_in_role(self, user_email):
        uri = REMOVE_USER_MEMBERSHIP_IN_ROLE_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            role_name=self._role_name,
            user_email=user_email,
        )
        hdrs = auth.set_header(
            self._auth, headers={'Accept': 'application/json', 'Content-Type': 'application/json'}
        )
        resp = requests.delete(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def verify_a_user_roles_permission_on_a_specific_RBAC_resource(self, permission, resource_path):
        uri = VERIFY_A_USER_ROLES_PERMISSION_ON_A_SPECIFIC_RBAC_RESOURCE_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            role_name=self._role_name,
            permission=permission,
            resource_path=resource_path,
        )
        hdrs = auth.set_header(
            self._auth,
            headers={'Accept': 'application/json', 'Content-Type': 'application/octet-stream'},
        )
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def verify_user_role_membership(self, user_email):
        uri = VERIFY_USER_ROLE_MEMBERSHIP_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            role_name=self._role_name,
            user_email=user_email,
        )
        hdrs = auth.set_header(
            self._auth, headers={'Accept': 'application/json', 'Content-Type': 'application/json'}
        )
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    @staticmethod
    def sort_permissions(resource_permissions):
        for i in range(len(resource_permissions.get('resourcePermission'))):
            resource_permissions['resourcePermission'][i]['permissions'].sort()
        return resource_permissions



import os
import re
import sys
# read the contents of your README file
from os import path

from apigee import APP
from apigee import __version__ as version
from apigee import description

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

SETUP_ARGS = dict(
    name=APP,
    version=version,
    description=(description),
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/darumatic/apigee-cli',
    author='Matthew Delotavo',
    author_email='matthew.t.delotavo@gmail.com',
    license='Apache license 2.0',
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
    ],
    # py_modules = ['apigee',],
    # entry_points='''
    #     [console_scripts]
    #     apigee=apigee.__main__:main
    # ''',
    entry_points={'console_scripts': ['apigee=apigee.__main__:main']},
    install_requires=[
        'requests>=2.22',
        'click',
        'click-aliases',
        'click-option-group',
        'colorama',
        'pyotp',
        'tqdm',
        'tabulate',
        'pyjwt',
        'python-gnupg>=0.3.5',
        'gitpython'
    ],
    project_urls={'Documentation': 'https://darumatic.github.io/apigee-cli/index.html'},
    python_requires='>=3.6',
)

if __name__ == '__main__':
    from setuptools import setup, find_packages

    SETUP_ARGS['packages'] = find_packages()
    setup(**SETUP_ARGS)



import click
from click_option_group import MutuallyExclusiveOptionGroup, optgroup

from apigee import console
from apigee.apps.apps import Apps
from apigee.auth import common_auth_options, gen_auth
# from apigee.cls import OptionEatAll
from apigee.prefix import common_prefix_options
from apigee.silent import common_silent_options
from apigee.verbose import common_verbose_options


@click.group(help='Management APIs available for working with developer apps.')
def apps():
    pass


def _create_developer_app(
    username, password, mfa_secret, token, zonename, org, profile, name, developer, body, **kwargs
):
    return (
        Apps(gen_auth(username, password, mfa_secret, token, zonename), org, None)
        .create_developer_app(developer, body)
        .text
    )


@apps.command(
    help='Creates an app associated with a developer, associates the app with an API product, and auto-generates an API key for the app to use in calls to API proxies inside the API product.'
)
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option('-n', '--name', help='name', required=True)
@click.option('--developer', help='developer email or id', required=True)
@click.option('-b', '--body', help='request body', required=True)
def create(*args, **kwargs):
    console.echo(_create_developer_app(*args, **kwargs))


def _delete_developer_app(
    username, password, mfa_secret, token, zonename, org, profile, name, developer, **kwargs
):
    return (
        Apps(gen_auth(username, password, mfa_secret, token, zonename), org, name)
        .delete_developer_app(developer)
        .text
    )


@apps.command(help='Deletes a developer app.')
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option('-n', '--name', help='name', required=True)
@click.option('--developer', help='developer email or id', required=True)
def delete(*args, **kwargs):
    console.echo(_delete_developer_app(*args, **kwargs))


def _create_empty_developer_app(
    username,
    password,
    mfa_secret,
    token,
    zonename,
    org,
    profile,
    name,
    developer,
    display_name="",
    callback_url="",
    **kwargs
):
    return (
        Apps(gen_auth(username, password, mfa_secret, token, zonename), org, name)
        .create_empty_developer_app(developer, display_name=display_name, callback_url=callback_url)
        .text
    )


@apps.command(help='Creates an empty developer app.')
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option('-n', '--name', help='name', required=True)
@click.option('--developer', help='developer email or id', required=True)
@click.option(
    '--display-name',
    default=None,
    help="The DisplayName (set with an attribute) is what appears in the management UI. If you don't provide a DisplayName, the name is used.",
)
@click.option(
    '--callback-url',
    default=None,
    help='The callbackUrl is used by OAuth 2.0 authorization servers to communicate authorization codes back to apps. CallbackUrl must match the value of redirect_uri in some OAuth 2.0 See the documentation on OAuth 2.0 for more details.',
)
def create_empty(*args, **kwargs):
    console.echo(_create_empty_developer_app(*args, **kwargs))


def _get_developer_app_details(
    username, password, mfa_secret, token, zonename, org, profile, name, developer, **kwargs
):
    return (
        Apps(gen_auth(username, password, mfa_secret, token, zonename), org, name)
        .get_developer_app_details(developer)
        .text
    )


@apps.command(
    help='Get the profile of a specific developer app. All times in the response are UNIX times. Note that the response contains a top-level attribute named accessType that is no longer used by Apigee.'
)
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option('-n', '--name', help='name', required=True)
@click.option('--developer', help='developer email or id', required=True)
def get(*args, **kwargs):
    console.echo(_get_developer_app_details(*args, **kwargs))


def _list_developer_apps(
    username,
    password,
    mfa_secret,
    token,
    zonename,
    org,
    profile,
    developer,
    prefix=None,
    expand=False,
    count=1000,
    startkey="",
    **kwargs
):
    return Apps(
        gen_auth(username, password, mfa_secret, token, zonename), org, None
    ).list_developer_apps(developer, prefix=prefix, expand=expand, count=count, startkey=startkey)


@apps.command(
    help="Lists all apps created by a developer in an organization, and optionally provides an expanded view of the apps. All time values in the response are UNIX times. You can specify either the developer's email address or Edge ID."
)
@common_auth_options
@common_verbose_options
@common_silent_options
@common_prefix_options
@click.option('--developer', help='developer email or id', required=True)
@click.option(
    '--expand/--no-expand',
    default=False,
    help='Set to true to expand the results. This query parameter does not work if you use the count or startKey query parameters.',
)
@click.option(
    '--count',
    type=click.INT,
    default=1000,
    show_default=True,
    help='Limits the list to the number you specify. The limit is 100. Use with the startKey parameter to provide more targeted filtering.',
)
@click.option(
    '--startkey',
    default="",
    show_default=True,
    help='To filter the keys that are returned, enter the name of a developer app that the list will start with.',
)
def list(*args, **kwargs):
    console.echo(_list_developer_apps(*args, **kwargs))


# def _list_apps_for_all_developers(
#     username, password, mfa_secret, token, zonename, org, profile,
#     prefix=None,
#     expand=False,
#     count=1000,
#     startkey="",
#     format="dict",
#     progress_bar=False,
# ):
#     pass

# def _delete_key_for_a_developer_app(username, password, mfa_secret, token, zonename, org, profile, developer, consumer_key):
#     pass


def _create_a_consumer_key_and_secret(
    username,
    password,
    mfa_secret,
    token,
    zonename,
    org,
    profile,
    name,
    developer,
    consumer_key=None,
    consumer_secret=None,
    key_length=32,
    secret_length=32,
    key_suffix=None,
    key_delimiter='-',
    products=[],
    **kwargs
):
    return (
        Apps(gen_auth(username, password, mfa_secret, token, zonename), org, name)
        .create_a_consumer_key_and_secret(
            developer,
            consumer_key=consumer_key,
            consumer_secret=consumer_secret,
            key_length=key_length,
            secret_length=secret_length,
            key_suffix=key_suffix,
            key_delimiter=key_delimiter,
            products=products,
        )
        .text
    )


@apps.command(
    help='Creates a custom consumer key and secret for a developer app. This is particularly useful if you want to migrate existing consumer keys/secrets to Edge from another system. Consumer keys and secrets can contain letters, numbers, underscores, and hyphens. No other special characters are allowed.'
)
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option('-n', '--name', help='app name', required=True)
@click.option('--developer', help='developer email or id', required=True)
@optgroup.group(
    'consumerKey options', cls=MutuallyExclusiveOptionGroup, help='The consumerKey options'
)
@optgroup.option('--consumer-key', default=None, help="")
@optgroup.option(
    '--key-length', type=click.INT, default=32, show_default=True, help='length of the consumer key'
)
@optgroup.group(
    'consumerSecret options', cls=MutuallyExclusiveOptionGroup, help='The consumerSecret options'
)
@optgroup.option('--consumer-secret', default=None, help="")
@optgroup.option(
    '--secret-length',
    type=click.INT,
    default=32,
    show_default=True,
    help='length of the consumer secret',
)
@click.option('--key-suffix', default=None, help="")
@click.option(
    '--key-delimiter', default='-', help='separates consumerKey and key suffix with a delimiter.'
)
@click.option(
    '--products',
    multiple=True,
    default=[],
    show_default=True,
    help="API products to be associated with the app's credentials",
)
# @click.option(
#     'products',
#     '--products',
#     metavar='LIST',
#     cls=OptionEatAll,
#     default=[],
#     help="A list of API products to be associated with the app's credentials",
# )
def create_creds(*args, **kwargs):
    # click.echo(kwargs.get('products'))
    # import sys;sys.exit(1)
    console.echo(_create_a_consumer_key_and_secret(*args, **kwargs))


# def _add_api_product_to_key(username, password, mfa_secret, token, zonename, org, profile, developer, consumer_key, request_body):
#     pass


def _restore_app(username, password, mfa_secret, token, zonename, org, profile, file, **kwargs):
    return (
        Apps(gen_auth(username, password, mfa_secret, token, zonename), org, None)
        .restore_app(file)
        .text
    )


@apps.command(help='Restore developer app from a file.')
@common_auth_options
@common_silent_options
@common_verbose_options
@click.option(
    '-f',
    '--file',
    type=click.Path(exists=True, dir_okay=False, file_okay=True, resolve_path=False),
    required=True,
)
def restore(*args, **kwargs):
    _restore_app(*args, **kwargs)



import click

from apigee import console
from apigee.apiproducts.apiproducts import Apiproducts
from apigee.auth import common_auth_options, gen_auth
from apigee.prefix import common_prefix_options
from apigee.silent import common_silent_options
from apigee.verbose import common_verbose_options


@click.group(
    help='API products enable you to bundle and distribute your APIs to multiple developer groups simultaneously, without having to modify code.'
)
def apiproducts():
    pass


def _create_api_product(
    username, password, mfa_secret, token, zonename, org, profile, name, body, **kwargs
):
    return (
        Apiproducts(gen_auth(username, password, mfa_secret, token, zonename), org, name)
        .create_api_product(body)
        .text
    )


@apiproducts.command(help='Creates an API product in an organization.')
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option('-n', '--name', help='name', required=True)
@click.option('-b', '--body', help='request body', required=True)
def create(*args, **kwargs):
    console.echo(_create_api_product(*args, **kwargs))


def _delete_api_product(
    username, password, mfa_secret, token, zonename, org, profile, name, **kwargs
):
    return (
        Apiproducts(gen_auth(username, password, mfa_secret, token, zonename), org, name)
        .delete_api_product()
        .text
    )


@apiproducts.command(help='Deletes an API product from an organization.')
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option('-n', '--name', help='name', required=True)
def delete(*args, **kwargs):
    console.echo(_delete_api_product(*args, **kwargs))


def _get_api_product(username, password, mfa_secret, token, zonename, org, profile, name, **kwargs):
    return (
        Apiproducts(gen_auth(username, password, mfa_secret, token, zonename), org, name)
        .get_api_product()
        .text
    )


@apiproducts.command(
    help='Gets configuration data for an API product. The API product name required in the request URL is not the "Display Name" value displayed for the API product in the Edge UI. While they may be the same, they are not always the same depending on whether the API product was created via UI or API.'
)
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option('-n', '--name', help='name', required=True)
def get(*args, **kwargs):
    console.echo(_get_api_product(*args, **kwargs))


def _list_api_products(
    username,
    password,
    mfa_secret,
    token,
    zonename,
    org,
    profile,
    prefix=None,
    expand=False,
    count=1000,
    startkey="",
    **kwargs
):
    return Apiproducts(
        gen_auth(username, password, mfa_secret, token, zonename), org, None
    ).list_api_products(prefix=prefix, expand=expand, count=count, startkey=startkey)


@apiproducts.command(help='Get a list of all API product names for an organization.')
@common_auth_options
@common_verbose_options
@common_silent_options
@common_prefix_options
@click.option(
    '--expand/--no-expand',
    default=False,
    help="Set to 'true' to get expanded details about each product.",
)
@click.option(
    '--count',
    type=click.INT,
    default=1000,
    show_default=True,
    help='Number of API products to return in the API call. The maximum limit is 1000. Use with the startkey to provide more targeted filtering.',
)
@click.option(
    '--startkey',
    default="",
    show_default=True,
    help='Returns a list of API products starting with the specified API product.',
)
def list(*args, **kwargs):
    console.echo(_list_api_products(*args, **kwargs))


def _update_api_product(
    username, password, mfa_secret, token, zonename, org, profile, body, **kwargs
):
    return (
        Apiproducts(gen_auth(username, password, mfa_secret, token, zonename), org, name)
        .update_api_product(body)
        .text
    )


@apiproducts.command(
    help='Updates an existing API product. You must include all required values, whether or not you are updating them, as well as any optional values that you are updating.'
)
@common_auth_options
@common_verbose_options
@common_silent_options
@click.option('-n', '--name', help='name', required=True)
@click.option('-b', '--body', help='request body', required=True)
def update(*args, **kwargs):
    console.echo(_update_api_product(*args, **kwargs))


def _push_apiproducts(
    username, password, mfa_secret, token, zonename, org, profile, file, **kwargs
):
    return Apiproducts(
        gen_auth(username, password, mfa_secret, token, zonename), org, None
    ).push_apiproducts(file)


@apiproducts.command(help='Push API product to Apigee. This will create/update an API product.')
@common_auth_options
@common_silent_options
@common_verbose_options
@click.option(
    '-f',
    '--file',
    type=click.Path(exists=True, dir_okay=False, file_okay=True, resolve_path=False),
    required=True,
)
def push(*args, **kwargs):
    _push_apiproducts(*args, **kwargs)



import builtins
import http.client as http_client
import logging

import click
import requests


def verbose_callback(ctx, param, value):
    builtins.APIGEE_CLI_TOGGLE_VERBOSE = value
    http_client.HTTPConnection.debuglevel = value
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    # requests_log = logging.getLogger('requests.packages.urllib3')
    # requests_log.setLevel(logging.DEBUG)
    # requests_log.propagate = True


def common_verbose_options(func):
    func = click.option(
        '-v',
        '--verbose',
        show_default='toggle verbose output',
        count=True,
        callback=verbose_callback,
    )(func)
    return func



import json
import sys

from apigee.crypto import (ENCRYPTED_HEADER_BEGIN, ENCRYPTED_HEADER_END,
                           decrypt_message, encrypt_message, is_encrypted)


class KeyvaluemapsSerializer:
    def diff_kvms(self, kvm1, kvm2):
        current_keys = [entry['name'] for entry in kvm1['entry']]
        deleted_keys = [entry for entry in kvm2['entry'] if entry['name'] not in current_keys]
        return current_keys, deleted_keys

    def encrypt_decrypt_keyvaluemap(self, kvm_dict, secret, encrypt=True):
        crypto_count = 0

        def _func(value, kvm, index, secret):
            operation = ''
            if encrypt:
                operation = 'encrypt_value' if not is_encrypted(value) else ''
            else:
                operation = 'decrypt_value' if is_encrypted(value) else ''
            if operation:
                getattr(KeyvaluemapsSerializer(), operation)(kvm, index, secret)
                return 1
            return 0

        if kvm_dict['encrypted'] and kvm_dict['entry']:
            for idx, entry in enumerate(kvm_dict['entry']):
                if entry.get('name') and entry.get('value'):
                    crypto_count += _func(kvm_dict['entry'][idx]['value'], kvm_dict, idx, secret)
        return kvm_dict, crypto_count

    def encrypt_value(self, kvm_dict, index, secret):
        plaintext = kvm_dict['entry'][index]['value']
        kvm_dict['entry'][index][
            'value'
        ] = f'{ENCRYPTED_HEADER_BEGIN}{encrypt_message(secret, plaintext)}{ENCRYPTED_HEADER_END}'

    def decrypt_value(self, kvm_dict, index, secret):
        ciphertext = kvm_dict['entry'][index]['value']
        decrypted = decrypt_message(secret, ciphertext)
        if decrypted == '':
            sys.exit('Incorrect symmetric key.')
        kvm_dict['entry'][index]['value'] = decrypted

    def serialize_details(self, maps, format, prefix=None):
        resp = maps
        if format == 'text':
            return maps.text
        maps = maps.json()
        if prefix:
            maps = [map for map in maps if map.startswith(prefix)]
        if format == 'json':
            return json.dumps(maps)
        elif format == 'table':
            pass
        elif format == 'dict':
            return maps
        return resp



import builtins
import functools
import inspect
import json
import logging
import sys

from apigee import console
from apigee.utils import remove_file_above_size


class InvalidApisError(Exception):
    pass


class NotYetImplementedError(Exception):
    pass


def setup_global_logger(log_file):
    logging.basicConfig(
        filename=log_file,
        level=logging.WARNING,
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
    )
    remove_file_above_size(log_file, size_kb=1000)


def exception_handler(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
            return result
        except Exception as e:
            logging.error('Exception occurred', exc_info=True)
            frm = inspect.trace()[-1]
            mod = inspect.getmodule(frm[0])
            modname = mod.__name__ if mod else frm[1]
            sys.exit(f'An exception of type {modname}.{type(e).__name__} occurred. Arguments:\n{e}')
        except KeyboardInterrupt:
            console.echo()
            sys.exit(130)

    return wrapper



import json

import requests
from requests.exceptions import HTTPError

from apigee import APIGEE_ADMIN_API_URL, auth, console
from apigee.targetservers.serializer import TargetserversSerializer
from apigee.utils import read_file

CREATE_A_TARGETSERVER_PATH = (
    '{api_url}/v1/organizations/{org}/environments/{environment}/targetservers'
)
DELETE_A_TARGETSERVER_PATH = (
    '{api_url}/v1/organizations/{org}/environments/{environment}/targetservers/{name}'
)
LIST_TARGETSERVERS_IN_AN_ENVIRONMENT_PATH = (
    '{api_url}/v1/organizations/{org}/environments/{environment}/targetservers'
)
GET_TARGETSERVER_PATH = (
    '{api_url}/v1/organizations/{org}/environments/{environment}/targetservers/{name}'
)
UPDATE_A_TARGETSERVER_PATH = (
    '{api_url}/v1/organizations/{org}/environments/{environment}/targetservers/{name}'
)


class Targetservers:
    def __init__(self, auth, org_name, targetserver_name):
        self._auth = auth
        self._org_name = org_name
        self._targetserver_name = targetserver_name

    def __call__(self):
        pass

    @property
    def auth(self):
        return self._auth

    @auth.setter
    def auth(self, value):
        self._auth = value

    @property
    def org_name(self):
        return self._org_name

    @org_name.setter
    def org_name(self, value):
        self._org_name = value

    @property
    def targetserver_name(self):
        return self._targetserver_name

    @targetserver_name.setter
    def targetserver_name(self, value):
        self._targetserver_name = value

    def create_a_targetserver(self, environment, request_body):
        uri = CREATE_A_TARGETSERVER_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, environment=environment
        )
        hdrs = auth.set_header(
            self._auth, headers={'Accept': 'application/json', 'Content-Type': 'application/json'}
        )
        body = json.loads(request_body)
        resp = requests.post(uri, headers=hdrs, json=body)
        resp.raise_for_status()
        return resp

    def delete_a_targetserver(self, environment):
        uri = DELETE_A_TARGETSERVER_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            environment=environment,
            name=self._targetserver_name,
        )
        hdrs = auth.set_header(self._auth, headers={'Accept': 'application/json'})
        resp = requests.delete(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def list_targetservers_in_an_environment(self, environment, prefix=None, format='json'):
        uri = LIST_TARGETSERVERS_IN_AN_ENVIRONMENT_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, environment=environment
        )
        hdrs = auth.set_header(self._auth, headers={'Accept': 'application/json'})
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return TargetserversSerializer().serialize_details(resp, format, prefix=prefix)

    def get_targetserver(self, environment):
        uri = GET_TARGETSERVER_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            environment=environment,
            name=self._targetserver_name,
        )
        hdrs = auth.set_header(self._auth, headers={'Accept': 'application/json'})
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def update_a_targetserver(self, environment, request_body):
        uri = UPDATE_A_TARGETSERVER_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            environment=environment,
            name=self._targetserver_name,
        )
        hdrs = auth.set_header(
            self._auth, headers={'Accept': 'application/json', 'Content-Type': 'application/json'}
        )
        body = json.loads(request_body)
        resp = requests.put(uri, headers=hdrs, json=body)
        resp.raise_for_status()
        return resp

    def push_targetserver(self, environment, file):
        targetserver = read_file(file, type='json')
        self._targetserver_name = targetserver['name']
        try:
            self.get_targetserver(environment)
            console.echo(f'Updating {self._targetserver_name}')
            console.echo(self.update_a_targetserver(environment, json.dumps(targetserver)).text)
        except HTTPError as e:
            if e.response.status_code != 404:
                raise e
            console.echo(f'Creating {self._targetserver_name}')
            console.echo(self.create_a_targetserver(environment, json.dumps(targetserver)).text)



import json
import logging
import os
import sys
import xml.etree.ElementTree as et
from pathlib import Path

import requests

from apigee import APIGEE_ADMIN_API_URL, auth, console
from apigee.apis.interfaces.apis_interface import InformalApisInterface
from apigee.apis.interfaces.pull_interface import InformalPullInterface
from apigee.apis.serializer import ApisSerializer
from apigee.caches.caches import Caches
from apigee.deployments.deployments import Deployments
from apigee.keyvaluemaps.keyvaluemaps import Keyvaluemaps
from apigee.targetservers.targetservers import Targetservers
from apigee.utils import (extract_zip, is_dir, make_dirs, path_exists,
                          paths_exist, run_func_on_dir_files,
                          run_func_on_iterable, split_path, write_zip)

DELETE_API_PROXY_REVISION_PATH = (
    '{api_url}/v1/organizations/{org}/apis/{api_name}/revisions/{revision_number}'
)
DEPLOY_API_PROXY_REVISION_PATH = '{api_url}/v1/organizations/{org}/environments/{environment}/apis/{api_name}/revisions/{revision_number}/deployments?delay={delay}'
EXPORT_API_PROXY_PATH = (
    '{api_url}/v1/organizations/{org}/apis/{api_name}/revisions/{revision_number}?format=bundle'
)
GET_API_PROXY_PATH = '{api_url}/v1/organizations/{org}/apis/{api_name}'
LIST_API_PROXIES_PATH = '{api_url}/v1/organizations/{org}/apis'
LIST_API_PROXY_REVISIONS_PATH = '{api_url}/v1/organizations/{org}/apis/{api_name}/revisions'
UNDEPLOY_API_PROXY_REVISION_PATH = '{api_url}/v1/organizations/{org}/environments/{environment}/apis/{api_name}/revisions/{revision_number}/deployments'
FORCE_UNDEPLOY_API_PROXY_REVISION_PATH = '{api_url}/v1/organizations/{org}/apis/{api_name}/revisions/{revision_number}/deployments?action=undeploy&env={environment}&force=true'


class Apis(InformalApisInterface, InformalPullInterface):
    def __init__(self, *args, **kwargs):
        if len(args) == 4:
            InformalPullInterface.__init__(self, args[0], args[1], args[2], args[3], **kwargs)
        else:
            InformalApisInterface.__init__(self, args[0], args[1])

    def __get_and_export(self, resource_type, files, environment, dependencies=[], force=True):
        resource = getattr(self, f'get_{resource_type}_dependencies')(files)
        dependencies.extend(resource)
        getattr(self, f'export_{resource_type}_dependencies')(environment, resource, force=force)
        return dependencies

    def delete_api_proxy_revision(self, api_name, revision_number):
        uri = DELETE_API_PROXY_REVISION_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            api_name=api_name,
            revision_number=revision_number,
        )
        hdrs = auth.set_header(self._auth, headers={'Accept': 'application/json'})
        resp = requests.delete(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def deploy_api_proxy_revision(
        self, api_name, environment, revision_number, delay=0, override=False
    ):
        uri = DEPLOY_API_PROXY_REVISION_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            environment=environment,
            api_name=api_name,
            revision_number=revision_number,
            delay=delay,
        )
        hdrs = auth.set_header(
            self._auth,
            headers={
                'Accept': 'application/json',
                'Content-Type': 'application/x-www-form-urlencoded',
            },
        )
        resp = requests.post(uri, headers=hdrs, data={'override': 'true' if override else 'false'})
        resp.raise_for_status()
        return resp

    def delete_undeployed_revisions(self, api_name, save_last=0, dry_run=False):
        details = ApisSerializer().filter_deployment_details(
            Deployments(self._auth, self._org_name, api_name)
            .get_api_proxy_deployment_details()
            .json()
        )
        undeployed = ApisSerializer().filter_undeployed_revisions(
            self.list_api_proxy_revisions(api_name).json(),
            ApisSerializer().filter_deployed_revisions(details),
            save_last=save_last,
        )
        console.echo(f'Undeployed revisions to be deleted: {undeployed}')
        if dry_run:
            return undeployed

        def _func(revision):
            console.echo(f'Deleting revison {revision}')
            self.delete_api_proxy_revision(api_name, revision)

        return run_func_on_iterable(undeployed, _func)

    def export_api_proxy(self, api_name, revision_number, fs_write=True, output_file=None):
        uri = EXPORT_API_PROXY_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            api_name=api_name,
            revision_number=revision_number,
        )
        hdrs = auth.set_header(self._auth, headers={'Accept': 'application/json'})
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        if fs_write:
            write_zip(output_file, resp.content)
        return resp

    def get_api_proxy(self, api_name):
        uri = GET_API_PROXY_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, api_name=api_name
        )
        hdrs = auth.set_header(self._auth, headers={'Accept': 'application/json'})
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def list_api_proxies(self, prefix=None, format='json'):
        uri = LIST_API_PROXIES_PATH.format(api_url=APIGEE_ADMIN_API_URL, org=self._org_name)
        hdrs = auth.set_header(self._auth, headers={'Accept': 'application/json'})
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return ApisSerializer().serialize_details(resp, format, prefix=prefix)

    def list_api_proxy_revisions(self, api_name):
        uri = LIST_API_PROXY_REVISIONS_PATH.format(
            api_url=APIGEE_ADMIN_API_URL, org=self._org_name, api_name=api_name
        )
        hdrs = auth.set_header(self._auth, headers={'Accept': 'application/json'})
        resp = requests.get(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def undeploy_api_proxy_revision(self, api_name, environment, revision_number):
        uri = UNDEPLOY_API_PROXY_REVISION_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            environment=environment,
            api_name=api_name,
            revision_number=revision_number,
        )
        hdrs = auth.set_header(self._auth, headers={'Accept': 'application/json'})
        resp = requests.delete(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def force_undeploy_api_proxy_revision(self, api_name, environment, revision_number):
        uri = FORCE_UNDEPLOY_API_PROXY_REVISION_PATH.format(
            api_url=APIGEE_ADMIN_API_URL,
            org=self._org_name,
            api_name=api_name,
            revision_number=revision_number,
            environment=environment,
        )
        hdrs = auth.set_header(self._auth, headers={'Accept': 'application/json'})
        resp = requests.delete(uri, headers=hdrs)
        resp.raise_for_status()
        return resp

    def get_apiproxy_files(self, directory):
        return run_func_on_dir_files(str(Path(directory) / 'apiproxy'), lambda f: f)

    def get_keyvaluemap_dependencies(self, files):
        def _func(f, _state):
            try:
                root = et.parse(f).getroot()
                if root.tag == 'KeyValueMapOperations':
                    if root.attrib['mapIdentifier'] not in _state:
                        _state.add(root.attrib['mapIdentifier'])
                        return root.attrib['mapIdentifier']
            except Exception as e:
                logging.warning(f'{e}; file={f}', exc_info=True)

        return run_func_on_iterable(files, _func, args=(set(),))

    def export_keyvaluemap_dependencies(
        self, environment, keyvaluemaps, force=False, expc_verbosity=1
    ):
        make_dirs(self._keyvaluemaps_dir)

        def _func(kvm):
            kvm_file = str(Path(self._keyvaluemaps_dir) / kvm)
            if not force:
                path_exists(os.path.relpath(kvm_file))
            resp = (
                Keyvaluemaps(self._auth, self._org_name, kvm)
                .get_keyvaluemap_in_an_environment(environment)
                .text
            )
            console.echo(resp, expc_verbosity=1)
            with open(kvm_file, 'w') as f:
                f.write(resp)

        return run_func_on_iterable(keyvaluemaps, _func)

    def get_targetserver_dependencies(self, files):
        def _func(f, _state):
            try:
                root = et.parse(f).getroot()
                for child in root.iter('Server'):
                    if child.attrib['name'] not in _state:
                        _state.add(child.attrib['name'])
                        return child.attrib['name']
            except Exception as e:
                logging.warning(f'{e}; file={f}', exc_info=True)

        return run_func_on_iterable(files, _func, args=(set(),))

    def export_targetserver_dependencies(
        self, environment, targetservers, force=False, expc_verbosity=1
    ):
        make_dirs(self._targetservers_dir)

        def _func(ts):
            ts_file = str(Path(self._targetservers_dir) / ts)
            if not force:
                path_exists(os.path.relpath(ts_file))
            resp = Targetservers(self._auth, self._org_name, ts).get_targetserver(environment).text
            console.echo(resp, expc_verbosity=1)
            with open(ts_file, 'w') as f:
                f.write(resp)

        return run_func_on_iterable(targetservers, _func)

    def get_cache_dependencies(self, files):
        def _func(f, _state):
            try:
                name = et.parse(f).find('.//CacheResource').text
                if name and name not in _state:
                    _state.add(name)
                    return name
            except Exception as e:
                logging.warning(f'{e}; file={f}', exc_info=True)

        return run_func_on_iterable(files, _func, args=(set(),))

    def export_cache_dependencies(self, environment, caches, force=False, expc_verbosity=1):
        make_dirs(self._caches_dir)

        def _func(cache):
            cache_file = str(Path(self._caches_dir) / cache)
            if not force:
                path_exists(os.path.relpath(cache_file))
            resp = (
                Caches(self._auth, self._org_name, cache)
                .get_information_about_a_cache(environment)
                .text
            )
            console.echo(resp, expc_verbosity=1)
            with open(cache_file, 'w') as f:
                f.write(resp)

        return run_func_on_iterable(caches, _func)

    def pull(self, api_name, dependencies=[], force=False, prefix=None, basepath=None):
        dependencies.append(api_name)
        make_dirs(self._work_tree)
        self.apiproxy_dir = api_name
        if not force:
            paths_exist((os.path.relpath(self._zip_file), os.path.relpath(self._apiproxy_dir)))
        export = self.export_api_proxy(
            api_name, self._revision_number, fs_write=True, output_file=self._zip_file
        )
        make_dirs(self._apiproxy_dir)
        extract_zip(self._zip_file, self._apiproxy_dir)
        os.remove(self._zip_file)
        files = self.get_apiproxy_files(self._apiproxy_dir)
        for resource_type in ('keyvaluemap', 'targetserver', 'cache'):
            self._Apis__get_and_export(
                resource_type, files, self._environment, dependencies=dependencies, force=force
            )
        return export, dependencies
